#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import sys

from .scaplatform import IS_WINDOWS


class _bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


bcolors = _bcolors()
color_support = True
if IS_WINDOWS:  # pragma: no cover
    try:
        # https://stackoverflow.com/questions/36760127/...
        # how-to-use-the-new-support-for-ansi-escape-sequences-in-the-windows-10-console
        from ctypes import windll  # type:ignore

        kernel32 = windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    except Exception:
        color_support = False


def color_print(
    color: str, s: str, prefix: str = "", postfix: str = ""
) -> None:  # pragma: no cover
    if color_support:
        sys.stderr.write(prefix)
        sys.stderr.write(bcolors.__getattribute__(color) + s + bcolors.ENDC)
        sys.stderr.write(postfix + "\n")
    else:  # pragma: no cover
        sys.stderr.write(prefix)
        sys.stderr.write(s)
        sys.stderr.write(postfix + "\n")


def same_line_print(s: str, width=80) -> None:  # pragma: no cover
    sys.stderr.write(f"\r{'':<{width}}")  # clear the line
    sys.stderr.write(f"\r{s}")


def get_yes_or_no(prompt: str = "") -> bool:  # pragma: no cover
    option = "Enter [y]es or [n]o: "
    sep = "\n" if prompt else ""
    answer = input(prompt + sep + option)
    while True:
        if answer in ("y", "Y", "yes", "Yes", "YES"):
            return True
        elif answer in ("n", "N", "no", "No", "NO"):
            return False
        else:
            answer = input(f"Invalid input: {answer}.\n{option}")



from collections import OrderedDict
from copy import deepcopy
import json
import logging
import os.path as os_path
from typing import Dict, List, Optional, Set, Union

from .scaexceptions import StructureNotFoundError


class Structure:
    def __init__(
        self,
        name: str,
        *,
        tregex_pattern: Optional[str] = None,
        dependency_pattern: Optional[str] = None,
        value_source: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """
        :param name: name of the structure
        :param tregex_pattern: Tregex pattern
        :param dependency_pattern: spaCy dependency pattern
        :param value_source: how to compute the value basing on values of other structures, e.g. "VP1 + VP2". One and only one of pattern and value_source should be given.
        :param description: description of the structure
        """
        self.name = name
        self.description = description

        # no need to check "W" because it uses regex
        if name != "W":
            count_non_none = sum(
                1
                for attr in (tregex_pattern, dependency_pattern, value_source)
                if attr is not None
            )
            if count_non_none != 1:
                raise ValueError(
                    "Exactly one of (tregex_pattern, dependency_pattern, value_source) should be"
                    " provided AND non-empty."
                )

        self.tregex_pattern = tregex_pattern
        self.dependency_pattern = dependency_pattern
        self.value_source = value_source

        self.value: Optional[Union[float, int]] = None
        self.matches: Optional[list] = None

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"name: {self.name}"
            + f"\ndescription: {self.description}"
            + f"\ntregex_pattern: {self.tregex_pattern}"
            + f"\nvalue_source: {self.value_source}"
            + f"\nmatches: {self.matches}"
            + f"\nfrequency: {self.value}"
        )

    def __add__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return self.value + other

        assert other.value is not None
        return self.value + other.value

    def __radd__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return other + self.value
        raise NotImplementedError()

    def __sub__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return self.value - other

        assert other.value is not None
        return self.value - other.value

    def __rsub__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return other - self.value
        raise NotImplementedError()

    def __mul__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return self.value * other

        assert other.value is not None
        return self.value * other.value

    def __rmul__(self, other) -> Union[int, float]:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return other * self.value
        raise NotImplementedError()

    def __truediv__(self, other) -> float:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return self.value / other if other else 0

        assert other.value is not None
        return self.value / other.value if other.value else 0

    def __rtruediv__(self, other) -> float:
        assert self.value is not None
        if isinstance(other, (float, int)):
            return other / self.value if self.value else 0
        raise NotImplementedError()


class StructureCounter:
    data_file = os_path.join(os_path.dirname(__file__), "data", "structure_data.json")
    with open(data_file, "r", encoding="utf-8") as f:
        BUILTIN_DATA = json.load(f)

    BUILTIN_STRUCTURE_DEFS: Dict[str, Structure] = {}
    for kwargs in BUILTIN_DATA["structures"]:
        BUILTIN_STRUCTURE_DEFS[kwargs["name"]] = Structure(**kwargs)

    DEFAULT_MEASURES: List[str] = [
        "W",
        "S",
        "VP",
        "C",
        "T",
        "DC",
        "CT",
        "CP",
        "CN",
        "MLS",
        "MLT",
        "MLC",
        "C/S",
        "VP/T",
        "C/T",
        "DC/C",
        "DC/T",
        "T/S",
        "CT/T",
        "CP/T",
        "CP/C",
        "CN/T",
        "CN/C",
    ]

    def __init__(
        self,
        ifile="",
        *,
        selected_measures: Optional[List[str]] = None,
        user_structure_defs: Optional[List[Dict[str, str]]] = None,
    ) -> None:
        self.ifile = ifile

        user_sname_structure_map: Dict[str, Structure] = {}
        user_snames: Optional[Set[str]] = None

        if user_structure_defs is not None:
            user_snames = StructureCounter.check_user_structure_def(user_structure_defs)
            logging.debug(f"[StructureCounter] user_snames: {user_snames}")

            for kwargs in user_structure_defs:
                user_sname_structure_map[kwargs["name"]] = Structure(**kwargs)

        self.sname_structure_map: Dict[str, Structure] = deepcopy(
            StructureCounter.BUILTIN_STRUCTURE_DEFS
        )
        self.sname_structure_map.update(user_sname_structure_map)

        default_measures = StructureCounter.DEFAULT_MEASURES + [
            sname
            for sname in user_sname_structure_map.keys()
            if sname not in StructureCounter.DEFAULT_MEASURES
        ]

        if selected_measures is not None:
            StructureCounter.check_undefined_measure(selected_measures, user_snames)
        self.selected_measures: List[str] = (
            selected_measures if selected_measures is not None else default_measures
        )
        logging.debug(f"[StructureCounter] selected_measures: {self.selected_measures}")

    @classmethod
    def check_user_structure_def(cls, user_structure_defs: List[Dict[str, str]]) -> Set[str]:
        """
        check duplicated definition, e.g., [{"name": "A", "tregex_pattern":"a"}, {"name": "A", "tregex_pattern":"a"}]
        check empty definition, e.g., [{"name": "A", "tregex_pattern":""}]
        """
        user_defined_snames = set()
        for definition in user_structure_defs:
            for k, v in definition.items():
                if len(v) == 0:
                    raise ValueError(f"Error! {k} is left empty.")
                if len(k) == 0:
                    raise ValueError(f"Error! {v} is assigned to empty attribute.")

            sname = definition["name"]
            if sname in user_defined_snames:
                raise ValueError(f'Duplicated structure definition "{sname}".')

            user_defined_snames.add(sname)
        logging.debug(f"[StructureCounter] user_defined_snames: {user_defined_snames}")
        return user_defined_snames

    @classmethod
    def check_undefined_measure(
        cls, selected_measures: List[str], user_defined_snames: Optional[Set[str]]
    ) -> None:
        # check undefined selected_measure
        if user_defined_snames is not None:
            all_measures = StructureCounter.BUILTIN_STRUCTURE_DEFS.keys() | user_defined_snames
        else:
            all_measures = set(StructureCounter.BUILTIN_STRUCTURE_DEFS.keys())
        logging.debug(f"[StructureCounter] all_measures: {all_measures}")

        for m in selected_measures:
            if m not in all_measures:
                raise ValueError(f"{m} has not been defined.")

    def get_structure(self, structure_name: str) -> Structure:
        try:
            structure = self.sname_structure_map[structure_name]
        except KeyError:
            raise StructureNotFoundError(f"{structure_name} not found.")
        else:
            return structure

    def set_matches(self, structure_name: str, matches: list) -> None:
        if structure_name not in self.sname_structure_map:
            raise StructureNotFoundError(f"{structure_name} not found")
        elif not isinstance(matches, list):
            raise ValueError("matches should be a list object")
        else:
            self.sname_structure_map[structure_name].matches = matches

    def get_matches(self, sname: str) -> Optional[list]:
        s = self.get_structure(sname)
        return s.matches

    def set_value(self, sname: str, value: Union[int, float]) -> None:
        if sname not in self.sname_structure_map:
            raise ValueError(f"{sname} not counted")
        elif not isinstance(value, (float, int)):
            raise ValueError(f"value should be either a float or an integer, got {value}")
        else:
            self.sname_structure_map[sname].value = value

    def get_value(self, sname: str, precision: int = 4) -> Optional[Union[float, int]]:
        value = self.get_structure(sname).value
        return round(value, precision) if value is not None else value

    def get_all_values(self, precision: int = 4) -> dict:
        # TODO should store Filename in an extra metadata layer
        # https://articles.zsxq.com/id_wnw0w98lzgsq.html
        freq_dict = OrderedDict({"Filename": self.ifile})
        for sname in self.selected_measures:
            freq_dict[sname] = str(self.get_value(sname, precision))
        return freq_dict

    def __add__(self, other: "StructureCounter") -> "StructureCounter":
        logging.debug("[StructureCounter] Adding counters...")
        new_ifile = self.ifile + "+" + other.ifile if self.ifile else other.ifile
        new_selected_measures = list(
            dict.fromkeys(self.selected_measures + other.selected_measures)
        )
        new = StructureCounter(new_ifile, selected_measures=new_selected_measures)
        snames_defined_by_value_source: List[str] = []
        for sname, structure in new.sname_structure_map.items():
            # structures defined by value_source should be re-calculated after
            # adding up structures defined by tregex_pattern
            if structure.value_source is not None:
                logging.debug(
                    f"[StructureCounter] Skip {sname} which is defined by value_source."
                )
                snames_defined_by_value_source.append(sname)
                continue

            this_value = self.get_value(sname)
            that_value = other.get_value(sname)
            value = (this_value or 0) + (that_value or 0)
            new.set_value(sname, value)
            logging.debug(
                f"[StructureCounter] Added {sname}: {this_value} + {that_value} = {value}"
            )
        return new



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import os
import subprocess
from subprocess import CalledProcessError
import sys

from .scaenv import getenv, STANFORD_TREGEX_HOME


def tregex_cli():
    stanford_tregex_home = getenv(STANFORD_TREGEX_HOME)
    if stanford_tregex_home is None:
        print(
            f"Error: The environment variable {STANFORD_TREGEX_HOME} is not found or "
            "its value is not an existing directory."
        )
        sys.exit(1)

    command = [
        "java",
        "-mx100m",
        "-cp",
        f"{stanford_tregex_home}{os.path.sep}stanford-tregex.jar{os.pathsep}",
        "edu.stanford.nlp.trees.tregex.TregexPattern",
    ] + sys.argv[1:]

    try:
        subprocess.run(command, check=True, capture_output=False)
    except CalledProcessError as e:
        print(e)
        sys.exit(1)


if __name__ == "__main__":
    tregex_cli()



#!/usr/bin/env python3
# -*- coding=utf-8 -*-


class CircularDefinitionError(ValueError):
    pass


class InvalidSourceError(ValueError):
    pass


class StructureNotFoundError(ValueError):
    pass



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import glob
import logging
import os
import os.path as os_path
from typing import List, Optional

from .scaplatform import IS_MAC, IS_LINUX, IS_WINDOWS
from .scaprint import color_print

JAVA_HOME = "JAVA_HOME"
STANFORD_PARSER_HOME = "STANFORD_PARSER_HOME"
STANFORD_TREGEX_HOME = "STANFORD_TREGEX_HOME"


def _setenv_windows(env_var: str, paths: List[str], is_refresh: bool = False) -> None:
    import winreg  # Allows access to the windows registry
    import ctypes  # Allows interface with low-level C API's

    with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # type:ignore
        # Get the current user registry
        with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # type:ignore
            # Go to the environment key
            if is_refresh or os.environ.get(env_var) is None:
                new_value = ";".join(paths)
            else:
                existing_value = os.environ.get(env_var)
                new_value = existing_value + ";" + ";".join(paths)  # type:ignore
                # Takes the current path value and appends the new program path
            winreg.SetValueEx(key, env_var, 0, winreg.REG_EXPAND_SZ, new_value)  # type:ignore
            # Updated the path with the updated path

        # Tell other processes to update their environment
        HWND_BROADCAST = 0xFFFF
        WM_SETTINGCHANGE = 0x1A
        SMTO_ABORTIFHUNG = 0x0002
        result = ctypes.c_long()
        SendMessageTimeoutW = ctypes.windll.user32.SendMessageTimeoutW  # type:ignore
        SendMessageTimeoutW(
            HWND_BROADCAST,
            WM_SETTINGCHANGE,
            0,
            "Environment",
            SMTO_ABORTIFHUNG,
            5000,
            ctypes.byref(result),
        )


def _setenv_unix(env_var: str, paths: List[str], is_refresh: bool = False) -> None:
    shell = os.environ.get("SHELL")
    if shell is None:
        logging.warning(
            "Failed to permanently append {path} to {env_var}.\nReason: can't detect current"
            " shell."
        )
    else:
        shell_rcfile = {
            "bash": "~/.bash_profile" if IS_MAC else "~/.bashrc",
            "zsh": "~/.zshrc",
            "ksh": "~/.kshrc",
            "tcsh": "~/.tcshrc",
            "csh": "~/.cshrc",
            "yash": "~/.yashrc",
            "fish": "~/.config/fish/config.fish",
            "ion": "~/.config/ion/initrc",
        }
        rcfile = shell_rcfile.get(os_path.basename(shell), None)
        if rcfile is None:
            logging.warning(
                "Failed to permanently set environment variables.\nReason: can't detect rc"
                f" file for {shell}."
            )
        else:
            new_paths = '"' + '":"'.join(paths) + '"'
            rcfile = os_path.expanduser(rcfile)
            if not os_path.isfile(rcfile):
                configs = []
            else:
                with open(rcfile, "r", encoding="utf-8") as f:
                    configs = [line.strip() for line in f.readlines()]
            new_config = (
                f"export {env_var}={new_paths}"
                if is_refresh
                else f"export {env_var}=${env_var}:{new_paths}"
            )
            duplicated_config_index = []
            for i, config in enumerate(configs):
                if is_refresh:
                    if config.startswith(f"export {env_var}"):
                        duplicated_config_index.append(i)
                else:
                    if config == new_config:
                        duplicated_config_index.append(i)
            for i in duplicated_config_index:
                del configs[i]
            if duplicated_config_index:
                configs.insert(duplicated_config_index[0], new_config)
            else:
                configs.append(new_config)
            with open(rcfile, "w", encoding="utf-8") as f:
                f.write("\n".join(configs))


def setenv(
    env_var: str, paths: List[str], is_override: bool = False, is_quiet: bool = False
) -> None:
    assert any((IS_WINDOWS, IS_MAC, IS_LINUX))
    if IS_WINDOWS:
        _setenv_windows(env_var, paths, is_override)
    else:
        _setenv_unix(env_var, paths, is_override)
    if not is_quiet:
        color_print(
            "OKGREEN",
            env_var,
            prefix="Added the following path(s) to ",
            postfix=":\n" + "\n".join(paths),
        )


def getenv(env_var: str) -> Optional[str]:
    directory = os.getenv(env_var, "")
    return directory if os_path.isdir(directory) else None


def search_java_home() -> Optional[str]:
    candidate = None
    paths = os.getenv("PATH", "").split(os.pathsep)
    for dir_name in paths:
        if os_path.basename(dir_name) == "bin":
            if glob.glob(os_path.join(dir_name, "java.*")):
                candidate = os_path.dirname(dir_name)
                break
    if candidate is None and IS_WINDOWS:
        system_software_dir = os.getenv("ProgramFiles", "")
        if glob.glob(os_path.join(system_software_dir, "Java", "j[dr][ke]*")):
            candidate = glob.glob(os_path.join(system_software_dir, "Java", "j[dr][ke]*"))[0]
    return candidate


def unite_classpaths(stanford_parser_home, stanford_tregex_home) -> list:
    cp_parser = os_path.join(stanford_parser_home, "*")
    cp_tregex = os_path.join(stanford_tregex_home, "stanford-tregex.jar")
    return [cp_parser, cp_tregex]



from collections import namedtuple
import email.message
import glob
import logging
import lzma
import os
import os.path as os_path
import re
import shutil
import subprocess
from sys import maxsize, platform
import tarfile
import tempfile
from typing import Optional, Union
from urllib.error import URLError
import urllib.parse
import urllib.request
import zipfile

from .scaplatform import IS_MAC, IS_WINDOWS, USER_SOFTWARE_DIR
from .scaprint import get_yes_or_no, same_line_print
from .util import SCAProcedureResult

if IS_MAC:
    import ssl

    try:
        _create_unverified_https_context = ssl._create_unverified_context
    except AttributeError:
        pass
    else:
        ssl._create_default_https_context = _create_unverified_https_context

_UNPACK200 = "unpack200.exe" if IS_WINDOWS else "unpack200"
_UNPACK200_ARGS = '-r -v -l ""' if IS_WINDOWS else ""

OS = "windows" if IS_WINDOWS else "mac" if IS_MAC else platform
ARCH = "x64" if maxsize > 2**32 else "x32"

_TAR = ".tar"
_TAR_GZ = ".tar.gz"
_ZIP = ".zip"
_SEVEN_ZIP = ".7z"

_Path = namedtuple("_Path", "dir base name ext")
JAVA = "Java"
_JAVA_VERSION = "8"
STANFORD_PARSER = "Stanford Parser"
STANFORD_TREGEX = "Stanford Tregex"


class Implementation:
    OPENJ9 = "openj9"
    HOTSPOT = "hotspot"


class DependsInstaller:
    def __init__(self) -> None:
        self._URL_JAVA_TEMPLATE = (
            "https://api.adoptopenjdk.net/v3/binary/latest/"
            "{}/ga/{}/{}/jdk/{}/normal/adoptopenjdk"
        )
        self._URL_JAVA_TEMPLATE_CHINA = (
            "https://mirrors.tuna.tsinghua.edu.cn/Adoptium/{}/jdk/{}/{}/"
        )
        self._URL_STANFORD_PARSER = (
            "https://downloads.cs.stanford.edu/nlp/software/stanford-parser-4.2.0.zip"
        )
        self._URL_STANFORD_TREGEX = (
            "https://downloads.cs.stanford.edu/nlp/software/stanford-tregex-4.2.0.zip"
        )
        self.headers = {"User-Agent": "Mozilla/5.0"}

    def normalize_version(self, version: str) -> str:
        if re.search(r"^\d+$", version):
            return version
        match = re.search(r"^\d+\.(\d+)$", version)
        if match:
            return match.group(1)
        raise ValueError(f"Error: Unexpected version format: {version}.")

    def get_java_download_url(
        self,
        version: str,
        operating_system: Optional[str] = None,
        arch: Optional[str] = None,
        impl: str = Implementation.HOTSPOT,
    ) -> str:
        version = self.normalize_version(version)
        self.is_use_chinese_jdk_mirror = get_yes_or_no(
            "Do you want to download Java from a Chinese mirror site? If you are inside of"
            " China, you may want to use this for a faster network connection."
        )
        if not self.is_use_chinese_jdk_mirror:
            return self._URL_JAVA_TEMPLATE.format(version, operating_system, arch, impl)
        else:
            index_url = self._URL_JAVA_TEMPLATE_CHINA.format(version, arch, operating_system)
            req = urllib.request.Request(index_url, headers=self.headers)
            response = urllib.request.urlopen(req)
            content = response.read().decode("utf-8")
            match = re.search(r'"([^"]+\.(?:zip|tar\.gz|tar|7z))"', content)
            if match:
                filename = match.group(1)
                return urllib.parse.urljoin(index_url, filename)
            else:
                raise URLError(
                    "Failed to find any archive file (*.zip, *.tar.gz, *.tar, or *.7z) at"
                    f" {index_url}."
                )

    def _get_normalized_archive_ext(self, file: str) -> str:
        if file.endswith(_TAR):
            return _TAR
        elif file.endswith(_TAR_GZ):
            return _TAR_GZ
        elif file.endswith(_ZIP):
            return _ZIP
        elif file.endswith(_SEVEN_ZIP):
            return _SEVEN_ZIP
        else:
            raise ValueError(f"Error: {file} has unexpected extension.")

    def _extract_files(self, file: str, file_ending: str, destination_folder: str) -> str:
        if not os_path.isfile(file):
            raise ValueError(f"Error: {file} is not a regular file.")

        start_listing = set(os.listdir(destination_folder))

        if file_ending == _TAR:
            with tarfile.open(file, "r:") as tar:
                tar.extractall(path=destination_folder)
        elif file_ending == _TAR_GZ:
            with tarfile.open(file, "r:gz") as tar:
                tar.extractall(path=destination_folder)
        elif file_ending == _ZIP:
            with zipfile.ZipFile(file, "r") as z:
                z.extractall(path=destination_folder)
        elif file_ending == _SEVEN_ZIP:
            with lzma.open(file, "rb") as z:
                with open(destination_folder, "wb") as destination:
                    shutil.copyfileobj(z, destination)

        end_listing = set(os.listdir(destination_folder))
        unzipped_directory = end_listing.difference(start_listing).pop()

        return os_path.join(destination_folder, unzipped_directory)

    def _path_parse(self, file_path: str) -> _Path:
        dirname = os_path.dirname(file_path)
        base = os_path.basename(file_path)
        name, ext = os_path.splitext(base)
        return _Path(dir=dirname, base=base, name=name, ext=ext)

    def _unpack_jars(self, fs_path: str, java_bin_path: str) -> None:
        if os_path.isdir(fs_path):
            for f in os.listdir(fs_path):
                current_path = os_path.join(fs_path, f)
                self._unpack_jars(current_path, java_bin_path)
            return
        elif os_path.isfile(fs_path):
            file_ext = os_path.splitext(fs_path)[-1]
            if file_ext.endswith("pack"):
                p = self._path_parse(fs_path)
                name = os_path.join(p.dir, p.name)
                tool_path = os_path.join(java_bin_path, _UNPACK200)
                try:
                    subprocess.run(
                        [tool_path, _UNPACK200_ARGS, f"{name}.pack", f"{name}.jar"],
                        check=True,
                        capture_output=True,
                    )
                except (FileNotFoundError, subprocess.CalledProcessError) as e:
                    raise RuntimeError(str(e))
            return
        else:
            raise ValueError(f"Error: {fs_path} is neither a directory not a file.")

    def _decompress_archive(
        self, archive_path: str, file_extension: str, target_dir: str
    ) -> str:
        logging.info(f"Decompressing {archive_path} to {target_dir}...")
        if not os_path.isdir(target_dir):
            os.makedirs(target_dir)

        archive_path = os_path.normpath(archive_path)

        if os_path.isfile(archive_path):
            unzipped_directory = self._extract_files(archive_path, file_extension, target_dir)
            return unzipped_directory
        elif os_path.isdir(archive_path):
            return archive_path
        else:
            raise ValueError(f"Error: {archive_path} is neither a directory not a file.")

    def _get_java_filename(self, download_url) -> str:
        if self.is_use_chinese_jdk_mirror:
            filename = urllib.parse.urlparse(download_url).path.rpartition("/")[-1]
        else:
            req = urllib.request.Request(download_url, headers=self.headers)
            response = urllib.request.urlopen(req)
            info = response.info()
            download_url = response.geturl()
            if "Content-Disposition" not in info:
                raise URLError(
                    f"Parsing the response from {download_url} failed. \nReason:"
                    ' "Content-Disposition" not in response.info().'
                )
            m = email.message.Message()
            m["content-type"] = info["Content-Disposition"]
            if m.get_param("filename") is None:
                raise URLError(
                    f"Parsing the response from {download_url} failed.\nReason: can't detect"
                    " the filename."
                )
            filename = m.get_param("filename")
            if not isinstance(filename, str):
                raise URLError(
                    f"Parsing the response from {download_url} failed.\nReason: the value of"
                    ' the attribute "filename" is not a str.'
                )
        return filename

    def _format_bytes(self, size: Union[int, float], precesion: int = 2):
        power = 2**10  # 1024
        n = 0
        power_labels = {0: "", 1: "K", 2: "M", 3: "G", 4: "T"}
        while size > power:
            size /= power
            n += 1
        return round(size, precesion), f"{power_labels[n]}"

    def _callbackfunc(self, block_num: int, block_size: int, total_size):
        max_euqal_sign_num = 50  # print up to 50 hashes
        precesion = 2
        downloaded_size = block_num * block_size
        if downloaded_size >= total_size:
            downloaded_size = total_size
        percent = int(100 * downloaded_size / total_size)
        downloaded_size, downloaded_size_unit = self._format_bytes(downloaded_size, precesion)
        total_size, total_size_unit = self._format_bytes(total_size, precesion)
        equal_sign_num = int(percent / 100 * max_euqal_sign_num)
        s = (
            f"{percent:3}% [{'=' * equal_sign_num}>{' '*(max_euqal_sign_num-equal_sign_num-1)}]"
            f" {downloaded_size:6} {downloaded_size_unit}/{total_size} {total_size_unit}"
        )
        same_line_print(s, width=100)

    def download(self, url: str, target_basename: Optional[str] = None) -> str:
        logging.info(f"Downloading {url}...")
        if target_basename is None:
            target_basename = urllib.parse.urlparse(url).path.rpartition("/")[-1]
            # e.g. stanford-tregex-4.2.0.zip, stanford-parser-4.2.0.zip
        target_path = os_path.join(tempfile.gettempdir(), target_basename)  # type: ignore

        try:
            opener = urllib.request.build_opener()
            opener.addheaders = list(self.headers.items())
            urllib.request.install_opener(opener)
            urllib.request.urlretrieve(url, target_path, self._callbackfunc)
            same_line_print("", width=100)
        except URLError as e:
            if hasattr(e, "reason"):
                raise URLError(f"Requesting to {url} failed.\nReason: {e.reason}")
            elif hasattr(e, "code"):
                raise URLError(f"Requesting to {url} failed.\nReason: {e.code}")
            else:
                raise URLError(f"Requesting to {url} failed.")

        return target_path

    def ask_install(self, name: str, is_assume_yes: bool = False) -> bool:
        reason_dict = {
            JAVA: f"values of PATH does not include a {JAVA} bin folder",
            STANFORD_PARSER: (
                "the environment variable STANFORD_PARSER_HOME is not found or its value is not"
                " an existing directory"
            ),
            STANFORD_TREGEX: (
                "the environment variable STANFORD_TREGEX_HOME is not found or its value is not"
                " an existing directory"
            ),
        }
        return get_yes_or_no(
            f"It seems that {name} has not been installed, because {reason_dict[name]}. Do"
            " you want to let NeoSCA install it for you?"
        )

    def install_java(
        self,
        version: str,
        operating_system: str,
        arch: str,
        impl: str,
        target_dir: str,
        is_assume_yes: bool = False,
    ) -> SCAProcedureResult:
        match = glob.glob(f"{target_dir}{os.sep}j[dr][ke]{version}*")
        if match:
            return True, match[0]
        is_install = True if is_assume_yes else self.ask_install(JAVA, is_assume_yes)
        if not is_install:
            return (
                False,
                (
                    f"You will have to install {JAVA} manually.\n\n1. To install it, visit"
                    " https://www.java.com/en/download/manual.jsp. Note that when mixing 64 bit"
                    " Python with 32 bit Java or vice versa the program will crash, so make"
                    " sure that you install a Java with the same bitness of your Python.\n2."
                    " Set an environment variable JAVA_HOME to the path of the unzipped"
                    " directory."
                ),
            )
        url = self.get_java_download_url(version, operating_system, arch, impl)
        target_basename = self._get_java_filename(url)  # e.g., jdk-18.0.2
        jdk_archive = self.download(url, target_basename)

        jdk_ext = self._get_normalized_archive_ext(jdk_archive)
        jdk_dir = self._decompress_archive(jdk_archive, jdk_ext, target_dir)
        jdk_bin = os_path.join(jdk_dir, "bin")
        self._unpack_jars(jdk_dir, jdk_bin)
        if jdk_archive:
            os.remove(jdk_archive)
        return True, jdk_dir

    def install_stanford(
        self, name: str, url: str, target_dir: str, is_assume_yes: bool = False
    ) -> SCAProcedureResult:
        match = glob.glob(f"{target_dir}{os.sep}{name.lower().replace(' ', '-')}*")
        if match:
            return True, match[0]
        is_install = True if is_assume_yes else self.ask_install(name, is_assume_yes)
        if not is_install:
            manual_install_prompt_dict = {
                STANFORD_PARSER: (
                    f"You will have to install {STANFORD_PARSER} manually.\n\n1. To install it,"
                    f" download and unzip the archive file at {self._URL_STANFORD_PARSER}.\n2."
                    " Set an environment variable STANFORD_PARSER_HOME to the path of the"
                    " unzipped directory."
                ),
                STANFORD_TREGEX: (
                    f"You will have to install {STANFORD_TREGEX} manually.\n\n1. To install it,"
                    f" download and unzip the archive file at {self._URL_STANFORD_TREGEX}.\n2."
                    " Set an environment variable STANFORD_PARSER_HOME to the path of the"
                    " unzipped directory."
                ),
            }
            return False, manual_install_prompt_dict[name]
        archive_file = self.download(url)
        archive_ext = self._get_normalized_archive_ext(archive_file)
        unzipped_directory = self._decompress_archive(archive_file, archive_ext, target_dir)
        if archive_file:
            os.remove(archive_file)
        return True, unzipped_directory

    def install(
        self,
        name: str,
        is_assume_yes: bool = False,
        version: str = _JAVA_VERSION,
        operating_system: str = OS,
        arch: str = ARCH,
        impl: str = Implementation.HOTSPOT,
        target_dir: str = USER_SOFTWARE_DIR,  # type: ignore
    ) -> SCAProcedureResult:
        if name == JAVA:
            return self.install_java(
                version, operating_system, arch, impl, target_dir, is_assume_yes
            )
        elif name == STANFORD_PARSER:
            return self.install_stanford(
                name, self._URL_STANFORD_PARSER, target_dir, is_assume_yes
            )
        elif name == STANFORD_TREGEX:
            return self.install_stanford(
                name, self._URL_STANFORD_TREGEX, target_dir, is_assume_yes
            )
        else:
            return False, f"Unexpected name: {name}."



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

from io import BytesIO
import logging
import os
import os.path as os_path
import re
import sys
from tokenize import NAME, NUMBER, PLUS, tokenize, untokenize
from typing import List, Optional, TYPE_CHECKING

import jpype
from jpype import JClass

from .scaexceptions import CircularDefinitionError, InvalidSourceError

if TYPE_CHECKING:
    from .structure_counter import StructureCounter


class StanfordTregex:
    def __init__(
        self,
        classpaths: Optional[list] = None,
        max_memory: str = "3072m",
    ) -> None:
        self.classpaths = classpaths if classpaths is not None else []
        self.max_memory = max_memory
        self.TREGEX_PATTERN = "edu.stanford.nlp.trees.tregex.TregexPattern"
        self.STRING_READER = "java.io.StringReader"
        self.PENN_TREE_READER = "edu.stanford.nlp.trees.PennTreeReader"
        self.init_tregex()

    def init_tregex(self):
        if not jpype.isJVMStarted():  # pragma: no cover
            logging.debug("[StanfordTregex] starting JVM...")
            # Note that isJVMStarted may be renamed to isJVMRunning in the future.
            # In jpype's _core.py:
            # > TODO This method is horribly named.  It should be named isJVMRunning as
            # > isJVMStarted would seem to imply that the JVM was started at some
            # > point without regard to whether it has been shutdown.
            jpype.startJVM(f"-Xmx{self.max_memory}", classpath=self.classpaths)

        self.TregexPattern = jpype.JClass(self.TREGEX_PATTERN)
        self.StringReader = JClass(self.STRING_READER)
        self.PennTreeReader = JClass(self.PENN_TREE_READER)
        self.compiled_pattern_map = {}

    def get_compiled_pattern(self, sname: str, pattern_string: str):
        if sname not in self.compiled_pattern_map:
            tregex_pattern = self.TregexPattern.compile(pattern_string)
            self.compiled_pattern_map[sname] = tregex_pattern
        else:
            tregex_pattern = self.compiled_pattern_map[sname]

        return tregex_pattern

    def get_matches(self, sname: str, pattern_string: str, trees: str) -> list:
        matches = []
        tregex_pattern = self.get_compiled_pattern(sname, pattern_string)

        treeReader = self.PennTreeReader(self.StringReader(trees))
        tree = treeReader.readTree()
        while tree is not None:
            matcher = tregex_pattern.matcher(tree)
            last_matching_root_node = None
            while matcher.find():
                match = matcher.getMatch()
                # Each tree node can be reported only once as the root of a match.
                # Although solely nodeNumber checking is enough, it involves
                #  iteration acorss the tree, which can be slow on high trees,
                #  so use equals() to exit if-condition early. The equals()
                #  will check labels, number of children, and finally whether
                #  the children are pairwise equal. This achieves an average
                #  speed increase of 78ms across 4 trials on ~5% fragment of
                #  Penn Treebank
                #  (https://raw.githubusercontent.com/nltk/nltk_data/gh-pages/packages/corpora/treebank.zip).
                if last_matching_root_node is not None and (
                    last_matching_root_node.equals(match)
                    and last_matching_root_node.nodeNumber(tree) == match.nodeNumber(tree)
                ):
                    continue
                last_matching_root_node = match

                # we don't use match.spanString() because the output lacks
                # whitespace, e.g., "the media" becomes "themedia"
                span_string = " ".join(str(leaf.toString()) for leaf in match.getLeaves())
                penn_string = str(match.pennString().replaceAll("\r", ""))
                matches.append(span_string + "\n" + penn_string)
            tree = treeReader.readTree()
        return matches

    @classmethod
    def check_circular_def(
        cls, descendant_sname: str, ancestor_snames: List[str], counter: "StructureCounter"
    ) -> None:
        if descendant_sname in ancestor_snames:
            circular_definition = ", ".join(
                f"{upstream_sname} = {counter.get_structure(upstream_sname).value_source}"
                for upstream_sname in ancestor_snames
            )
            raise CircularDefinitionError(f"Circular definition: {circular_definition}")
        else:
            logging.debug(
                "[StanfordTregex] Circular definition check passed: descendant"
                f" {descendant_sname} not in ancestors {ancestor_snames}"
            )

    def tokenize_value_source(
        self,
        value_source: str,
        counter: "StructureCounter",
        sname: str,
        trees: str,
        ancestor_snames: List[str],
    ) -> list:
        tokens = []
        g = tokenize(BytesIO(value_source.encode("utf-8")).readline)
        next(g)  # skip the "utf-8" token
        for toknum, tokval, *_ in g:
            if toknum == NAME:
                ancestor_snames.append(sname)
                StanfordTregex.check_circular_def(tokval, ancestor_snames, counter)

                self.set_value(counter, tokval, trees, ancestor_snames)
                if self.has_tregex_pattern(counter, tokval):
                    ancestor_snames.clear()

                tokens.append((toknum, f"counter.get_structure('{tokval}')"))
            elif toknum == NUMBER:
                tokens.append((toknum, tokval))
            elif tokval in ("+", "-", "*", "/", "(", ")"):
                tokens.append((toknum, tokval))
            # constrain value_source as only NAMEs and numberic ops to assure security for `eval`
            elif tokval != "":
                raise InvalidSourceError(f'Unexpected token: "{tokval}"')
        # append "+ 0" to force tokens evaluated as num if value_source contains just name of another Structure
        tokens.extend(((PLUS, "+"), (NUMBER, "0")))
        return tokens

    def has_tregex_pattern(self, counter: "StructureCounter", sname: str) -> bool:
        return counter.get_structure(sname).tregex_pattern is not None

    def set_value_from_pattern(self, counter: "StructureCounter", sname: str, trees: str):
        structure = counter.get_structure(sname)
        tregex_pattern = structure.tregex_pattern
        assert tregex_pattern is not None

        logging.info(
            f" Searching for {sname}"
            + (f" ({structure.description})..." if structure.description is not None else "...")
        )
        logging.debug(f" Searching for {tregex_pattern}")
        matched_subtrees = self.get_matches(sname, tregex_pattern, trees)
        counter.set_matches(sname, matched_subtrees)
        counter.set_value(sname, len(matched_subtrees))

    def set_value_from_source(
        self, counter: "StructureCounter", sname: str, trees: str, ancestor_snames: List[str]
    ) -> None:
        structure = counter.get_structure(sname)
        value_source = structure.value_source
        assert value_source is not None, f"value_source for {sname} is None."

        logging.info(
            f" Calculating {sname} "
            + (f"({structure.description}) " if structure.description is not None else "")
            + f"= {value_source}..."
        )
        tokens = self.tokenize_value_source(value_source, counter, sname, trees, ancestor_snames)
        value = eval(untokenize(tokens))
        counter.set_value(sname, value)

    def set_value(
        self,
        counter: "StructureCounter",
        sname: str,
        trees: str,
        ancestor_snames: List[str] = [],
    ) -> None:
        value = counter.get_value(sname)
        if value is not None:
            logging.debug(
                f"[StanfordTregex] {sname} has already been set as {value}, skipping..."
            )
            return

        if sname == "W":
            logging.info(' Searching for "words"')
            value = len(re.findall(r"\([A-Z]+\$? [^()—–-]+\)", trees))
            counter.set_value(sname, value)
            return

        if self.has_tregex_pattern(counter, sname):
            self.set_value_from_pattern(counter, sname, trees)
        else:
            self.set_value_from_source(counter, sname, trees, ancestor_snames)

    def set_all_values(self, counter: "StructureCounter", trees: str) -> None:
        for sname in counter.selected_measures:
            self.set_value(counter, sname, trees)

    def query(
        self,
        counter: "StructureCounter",
        trees: str,
        is_reserve_matched: bool = False,
        odir_matched: str = "",
        is_stdout: bool = False,
    ) -> "StructureCounter":
        self.set_all_values(counter, trees)

        if is_reserve_matched:  # pragma: no cover
            self.write_match_output(counter, odir_matched, is_stdout)
        return counter

    def write_match_output(
        self, counter: "StructureCounter", odir_matched: str = "", is_stdout: bool = False
    ) -> None:  # pragma: no cover
        """
        Save Tregex's match output
        """
        bn_input = os_path.basename(counter.ifile)
        bn_input_noext = os_path.splitext(bn_input)[0]
        subodir_matched = os_path.join(odir_matched, bn_input_noext).strip()
        if not is_stdout:
            os.makedirs(subodir_matched, exist_ok=True)
        for sname, structure in counter.sname_structure_map.items():
            matches = structure.matches
            if matches is None or len(matches) == 0:
                continue

            meta_data = (
                f"# name: {structure.name}\n"
                + f"# description: {structure.description}\n"
                + f"# tregex_pattern: {structure.tregex_pattern}\n\n"
            )
            res = "\n".join(matches)
            # only accept alphanumeric chars, underscore, and hypen
            escaped_sname = re.sub(r"[^\w-]", "", sname.replace("/", "-per-"))
            matches_id = bn_input_noext + "-" + escaped_sname
            if not is_stdout:
                extension = ".matched"
                fn_match_output = os_path.join(subodir_matched, matches_id + extension)
                with open(fn_match_output, "w", encoding="utf-8") as f:
                    f.write(meta_data)
                    f.write(res)
            else:
                sys.stdout.write(matches_id + "\n")
                sys.stdout.write(meta_data)
                sys.stdout.write(res)



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

from typing import Optional, Tuple

# For all the procedures in SCAUI, return a tuple as the result
# The first element bool indicates whether the procedure succeeds
# The second element is the error message if it fails.
SCAProcedureResult = Tuple[bool, Optional[str]]



import json
import logging
import os
import os.path as os_path
import sys
from typing import Dict, List, Optional, Set, Tuple

from .parser import StanfordParser
from .querier import StanfordTregex
from .scaenv import unite_classpaths
from .scaio import SCAIO
from .structure_counter import StructureCounter


class NeoSCA:
    def __init__(
        self,
        ofile_freq: str = "result.csv",
        oformat_freq: str = "csv",
        stanford_parser_home: str = "",
        stanford_tregex_home: str = "",
        odir_matched: str = "",
        newline_break: str = "never",
        max_length: Optional[int] = None,
        selected_measures: Optional[List[str]] = None,
        is_reserve_parsed: bool = False,
        is_reserve_matched: bool = False,
        is_stdout: bool = False,
        is_skip_querying: bool = False,
        is_skip_parsing: bool = False,
        is_pretokenized: bool = False,
        config: Optional[str] = None,
    ) -> None:
        self.ofile_freq = ofile_freq
        self.oformat_freq = oformat_freq
        self.classpaths = unite_classpaths(stanford_parser_home, stanford_tregex_home)
        self.odir_matched = odir_matched
        self.newline_break = newline_break
        self.max_length = max_length
        self.selected_measures = selected_measures
        self.is_reserve_parsed = is_reserve_parsed
        self.is_reserve_matched = is_reserve_matched
        self.is_stdout = is_stdout
        self.is_skip_querying = is_skip_querying
        self.is_skip_parsing = is_skip_parsing
        self.is_pretokenized = is_pretokenized

        self.user_data, self.user_structure_defs, self.user_snames = self.load_user_config(
            config
        )
        logging.debug(f"[NeoSCA] user_snames: {self.user_snames}")

        if selected_measures is not None:
            StructureCounter.check_undefined_measure(selected_measures, self.user_snames)

        self.counters: List[StructureCounter] = []
        self.io = SCAIO()
        self.is_stanford_parser_initialized = False
        self.is_stanford_tregex_initialized = False

    def load_user_config(
        self, config: Optional[str]
    ) -> Tuple[dict, List[dict], Optional[Set[str]]]:
        user_data: dict = {}
        user_structure_defs: List[Dict[str, str]] = []
        user_snames: Optional[Set[str]] = None

        if config is not None:
            with open(config, "r", encoding="utf-8") as f:
                user_data = json.load(f)

            user_structure_defs = user_data["structures"]
            user_snames = StructureCounter.check_user_structure_def(user_structure_defs)

        return user_data, user_structure_defs, user_snames

    def ensure_stanford_tregex_initialized(self) -> None:
        if not self.is_stanford_tregex_initialized:
            self.tregex = StanfordTregex(classpaths=self.classpaths)
            self.is_stanford_tregex_initialized = True

    def ensure_stanford_parser_initialized(self) -> None:
        if not self.is_stanford_parser_initialized:
            self.parser = StanfordParser(classpaths=self.classpaths)
            self.is_stanford_parser_initialized = True

    def already_parsed(self, ofile_parsed: str, ifile: str) -> bool:
        has_been_parsed = False
        is_exist = os_path.exists(ofile_parsed)
        if is_exist:
            is_not_empty = os_path.getsize(ofile_parsed) > 0
            is_parsed_newer_than_input = os_path.getmtime(ofile_parsed) > os_path.getmtime(ifile)
            if is_not_empty and is_parsed_newer_than_input:
                has_been_parsed = True
        return has_been_parsed

    def query_against_trees(self, trees: str, counter: StructureCounter) -> StructureCounter:
        self.ensure_stanford_tregex_initialized()
        counter = self.tregex.query(
            counter,
            trees,
            is_reserve_matched=self.is_reserve_matched,
            odir_matched=self.odir_matched,
            is_stdout=self.is_stdout,
        )
        return counter

    def parse_text(self, text: str, ofile_parsed="cmdline_text.parsed") -> str:
        if self.is_skip_parsing:  # assume input as parse trees
            return text

        self.ensure_stanford_parser_initialized()
        trees = self.parser.parse(
            text,
            max_length=self.max_length,
            newline_break=self.newline_break,
            is_pretokenized=self.is_pretokenized,
            is_reserve_parsed=self.is_reserve_parsed,
            ofile_parsed=ofile_parsed,
            is_stdout=self.is_stdout,
        )
        return trees

    def parse_ifile(self, ifile: str) -> Optional[str]:
        if self.is_skip_parsing:
            # assume input as parse trees
            return self.io.read_txt(ifile, is_guess_encoding=False)

        ofile_parsed = os_path.splitext(ifile)[0] + ".parsed"
        has_been_parsed = self.already_parsed(ofile_parsed=ofile_parsed, ifile=ifile)
        if has_been_parsed:
            logging.info(
                f"Parsing skipped: {ofile_parsed} already"
                f" exists, and is non-empty and newer than {ifile}."
            )
            # parse file are always (1) plain text, and (2) of utf-8 encoding
            return self.io.read_txt(ofile_parsed, is_guess_encoding=False)

        text = self.io.read_file(ifile)
        if text is None:
            return None

        try:
            trees = self.parse_text(text, ofile_parsed)
        except KeyboardInterrupt:
            if os_path.exists(ofile_parsed):
                os.remove(ofile_parsed)
            sys.exit(1)
        else:
            return trees

    def run_on_text(self, text: str, ifile: str = "cmdline_text") -> None:
        trees: str = self.parse_text(text)

        if self.is_skip_querying:
            return

        counter = StructureCounter(
            ifile,
            selected_measures=self.selected_measures,
            user_structure_defs=self.user_structure_defs,
        )
        counter = self.query_against_trees(trees, counter)
        self.counters.append(counter)
        self.write_value_output()

    def parse_and_query_ifile(self, ifile: str) -> Optional[StructureCounter]:
        trees = self.parse_ifile(ifile)
        if trees is None:
            return None

        counter = StructureCounter(
            ifile,
            selected_measures=self.selected_measures,
            user_structure_defs=self.user_structure_defs,
        )
        return self.query_against_trees(trees, counter)

    def parse_ifiles(self, ifiles: List[str]):
        total = len(ifiles)
        for i, ifile in enumerate(ifiles, 1):
            logging.info(f'[NeoSCA] Processing "{ifile}" ({i}/{total})...')
            self.parse_ifile(ifile)

    def parse_subfiles_list(self, subfiles_list: List[List[str]]):
        for subfiles in subfiles_list:
            self.parse_ifiles(subfiles)

    def parse_and_query_ifiles(self, ifiles):
        total = len(ifiles)
        for i, ifile in enumerate(ifiles, 1):
            logging.info(f'[NeoSCA] Processing "{ifile}" ({i}/{total})...')
            counter = self.parse_and_query_ifile(ifile)
            if counter is None:
                continue
            self.counters.append(counter)

    def parse_and_query_subfiles_list(self, subfiles_list: List[List[str]]):
        for subfiles in subfiles_list:
            total = len(subfiles)
            parent_counter = StructureCounter(
                selected_measures=self.selected_measures,
                user_structure_defs=self.user_structure_defs,
            )

            for i, subfile in enumerate(subfiles, 1):
                logging.info(f'[NeoSCA] Processing "{subfile}" ({i}/{total})...')
                child_counter = self.parse_and_query_ifile(subfile)
                if child_counter is None:
                    continue
                parent_counter += child_counter

            self.tregex.set_all_values(parent_counter, "")
            self.counters.append(parent_counter)

    def run_on_ifiles(self, files: List[str] = [], subfiles_list: List[List[str]] = []) -> None:
        if self.is_skip_querying:
            self.parse_ifiles(files)
            self.parse_subfiles_list(subfiles_list)
            return

        self.parse_and_query_ifiles(files)
        self.parse_and_query_subfiles_list(subfiles_list)
        self.write_value_output()

    def write_value_output(self) -> None:
        logging.debug("[NeoSCA] Writting counts and/or frequencies...")

        counters = self.counters
        if len(counters) == 0:
            raise ValueError("empty counter list")

        oformat_freq = self.oformat_freq
        if oformat_freq not in ("csv", "json"):
            raise ValueError(f'oformat_freq {oformat_freq} not in ("csv", "json")')

        sname_value_maps: List[Dict[str, str]] = [
            counter.get_all_values() for counter in counters
        ]

        handle = (
            open(self.ofile_freq, "w", encoding="utf-8", newline="")
            if not self.is_stdout
            else sys.stdout
        )

        if oformat_freq == "csv":
            import csv

            fieldnames = sname_value_maps[0].keys()
            writer = csv.DictWriter(handle, fieldnames=fieldnames)

            writer.writeheader()
            writer.writerows(sname_value_maps)

        else:
            import json

            json.dump(sname_value_maps, handle)

        handle.close()



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

try:
    from xml.etree.cElementTree import XML, fromstring
except ImportError:
    from xml.etree.ElementTree import XML, fromstring
import glob
import logging
import os.path as os_path
import sys
from typing import ByteString, Callable, Dict, List, Optional, Union
import zipfile

from charset_normalizer import detect

from .util import SCAProcedureResult


class SCAIO:
    DOCX_NAMESPACE = "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}"
    DOCX_PARA = DOCX_NAMESPACE + "p"
    DOCX_TEXT = DOCX_NAMESPACE + "t"

    ODT_NAMESPACE = ".//{urn:oasis:names:tc:opendocument:xmlns:text:1.0}"
    ODT_PARA = ODT_NAMESPACE + "p"

    def __init__(self):
        self.ext_read_map: Dict[str, Callable] = {
            "txt": self.read_txt,
            "docx": self.read_docx,
            "odt": self.read_odt,
        }
        # .parsed files, along with other types of files, should be explicitly
        # checked and excluded, because they are text files and self.read_txt can
        # only exclude non-text files
        self.extensions_to_exclude: tuple = (
            "parsed",
            "prd",
            "csv",
            "tsv",
            "xml",
            "json",
            "md",
            "yml",
            "toml",
            "html",
            "htm",
            "svg",
            "cfg",
            "conf",
            "log",
            "png",
            "jpg",
            "gif",
            "ini",
            "rtf",
            "bat",
            "sh",
            "py",
            "r",
            "R",
            "h",
            "java",
            "cpp",
            "sql",
            "textile",
            "srt",
            "tex",
        )
        self.previous_encoding: str = "utf-8"

    def read_docx(self, path: str) -> str:
        """
        Take the path of a docx file as argument, return the text in unicode.
        This approach does not extract text from headers and footers.

        https://etienned.github.io/posts/extract-text-from-word-docx-simply/
        """
        with zipfile.ZipFile(path) as zip_file:
            xml_content = zip_file.read("word/document.xml")
        tree = XML(xml_content)

        paragraphs = []
        for paragraph in tree.iter(self.DOCX_PARA):
            text = "".join(node.text for node in paragraph.iter(self.DOCX_TEXT) if node.text)
            paragraphs.append(text)
        return "\n".join(paragraphs)

    def read_odt(self, path: str) -> str:
        with zipfile.ZipFile(path) as zip_file:
            xml_content = zip_file.read("content.xml")
        root = fromstring(xml_content)
        paragraphs = root.findall(self.ODT_PARA)
        return "\n".join("".join(node.itertext()) for node in paragraphs)

    def _read_txt(
        self, path: str, mode: str, encoding: Optional[str] = None
    ) -> Union[str, ByteString]:
        try:
            with open(path, mode=mode, encoding=encoding) as f:
                content = f.read()
        # input file existence has already been checked in main.py, here check
        # it again in case users remove input files during runtime
        except FileNotFoundError:
            logging.critical(f"{path} does not exist.")
            sys.exit(1)
        else:
            return content

    def read_txt(self, path: str, is_guess_encoding: bool = True) -> Optional[str]:
        if not is_guess_encoding:
            return self._read_txt(path, "r", "utf-8")  # type:ignore

        try:
            logging.info(f"Attempting to read {path} with {self.previous_encoding} encoding...")
            content = self._read_txt(path, "r", self.previous_encoding)  # type:ignore
        except UnicodeDecodeError:
            logging.info(f"Attempt failed. Reading {path} in binary mode...")
            bytes_ = self._read_txt(path, "rb")
            logging.info("Guessing the encoding of the byte string...")
            encoding = detect(bytes_)["encoding"]  # type:ignore

            if encoding is None:
                logging.warning(f"{path} is of unsupported file type. Skipped.")
                return None

            self.previous_encoding = encoding  # type:ignore
            logging.info(f"Decoding the byte string with {encoding} encoding...")
            content = bytes_.decode(encoding=encoding)  # type:ignore

        return content  # type:ignore

    def _is_to_exclude(self, path: str) -> bool:
        *_, ext = path.split(".")
        if ext in self.extensions_to_exclude:
            return True
        return False

    def read_file(self, path: str) -> Optional[str]:
        if self._is_to_exclude(path):
            logging.warning(f"[SCAIO] {path} does not appear to be an input file. Skipping.")
            return None

        *_, ext = path.split(".")
        if ext not in self.ext_read_map:
            # assume files with other extensions as text files; if not so,
            # read_txt() will fail and log them and return None
            ext = "txt"
        return self.ext_read_map[ext](path)  # type:ignore

    @classmethod
    def is_writable(cls, filename: str) -> SCAProcedureResult:
        """check whether files are opened by such other processes as WPS"""
        if not os_path.exists(filename):
            return True, None
        try:
            with open(filename, "w", encoding="utf-8"):
                pass
        except PermissionError:
            return (
                False,
                (
                    f"PermissionError: can not write to {filename}, because it is already in use"
                    " by another process.\n\n1. Ensure that {filename} is closed, or \n2."
                    " Specify another output filename through the `-o` option, e.g. nsca"
                    f" input.txt -o {filename.replace('.csv', '-2.csv')}"
                ),
            )
        else:
            return True, None

    def load_pickle_lzma_file(self, filepath: str) -> dict:
        import lzma
        import pickle

        with open(filepath, "rb") as f:
            data_pickle_lzma = f.read()

        data_pickle = lzma.decompress(data_pickle_lzma)
        data = pickle.loads(data_pickle)

        return data

    def get_verified_ifile_list(self, ifile_list: List[str]) -> List[str]:
        verified_ifile_list = []
        for path in ifile_list:
            if os_path.isfile(path):
                if self._is_to_exclude(path):
                    logging.warning(
                        f"[SCAIO] {path} does not appear to be an input file. Skipping."
                    )
                    continue
                logging.debug(f"[SCAIO] Adding {path} to input file list")
                verified_ifile_list.append(path)
            elif os_path.isdir(path):
                verified_ifile_list.extend(
                    path
                    for path in glob.glob(f"{path}{os_path.sep}*")
                    if not self._is_to_exclude(path)
                )
            elif glob.glob(path):
                verified_ifile_list.extend(glob.glob(path))
            else:
                logging.critical(f"No such file as\n\n{path}")
                sys.exit(1)
        return verified_ifile_list



import argparse
import logging
import os.path as os_path
import sys
from typing import Callable, List

from ..scaio import SCAIO
from ..scaprint import color_print, get_yes_or_no
from ..util import SCAProcedureResult
from .lca import LCA


class LCAUI:
    def __init__(self) -> None:
        self.args_parser: argparse.ArgumentParser = self.create_args_parser()
        self.options: argparse.Namespace = argparse.Namespace()
        self.scaio = SCAIO()

        self.is_spacy_initialized: bool = False

    def create_args_parser(self) -> argparse.ArgumentParser:
        args_parser = argparse.ArgumentParser(
            prog="nsca-lca", formatter_class=argparse.RawDescriptionHelpFormatter
        )
        args_parser.add_argument(
            "--wordlist",
            dest="wordlist",
            choices=("bnc", "anc"),
            default="bnc",
            help=(
                "Choose BNC or ANC (American National Corpus) wordlist for lexical"
                ' sophistication analysis. The default is "bnc".'
            ),
        )
        args_parser.add_argument(
            "--tagset",
            dest="tagset",
            choices=("ud", "ptb"),
            default="ud",
            help='Choose UD or PTB POS tagset for word classification. The default is "ud".',
        )
        args_parser.add_argument(
            "--output-file",
            "-o",
            metavar="<filename>",
            dest="ofile",
            default=None,
            help='Specify an output file. The default is "result.csv".',
        )
        args_parser.add_argument(
            "--stdout",
            dest="is_stdout",
            action="store_true",
            default=False,
            help="Write the output to the stdout instead of saving it to a file.",
        )
        args_parser.add_argument(
            "--text",
            "-t",
            metavar="<text>",
            default=None,
            help="Pass text through the command line.",
        )
        args_parser.add_argument(
            "--quiet",
            dest="is_quiet",
            action="store_true",
            default=False,
            help="Stop the program from printing anything except for final results.",
        )
        args_parser.add_argument(
            "--verbose",
            dest="is_verbose",
            action="store_true",
            default=False,
            help="Print detailed logging messages.",
        )
        return args_parser

    def parse_args(self, argv: List[str]) -> SCAProcedureResult:
        options, ifile_list = self.args_parser.parse_known_args(argv[1:])

        assert not (
            options.is_quiet and options.is_verbose
        ), "logging cannot be quiet and verbose at the same time"
        if options.is_quiet:
            logging.basicConfig(format="%(message)s", level=logging.CRITICAL)
        elif options.is_verbose:
            logging.basicConfig(format="%(message)s", level=logging.DEBUG)
        else:
            logging.basicConfig(format="%(message)s", level=logging.INFO)

        ofile = options.ofile
        if ofile is not None:
            ofile_ext = os_path.splitext(options.ofile)[-1].lstrip(".")
            if ofile_ext != "csv":
                return (
                    False,
                    f"The file extension {ofile_ext} is not supported. Please use csv.",
                )
        else:
            options.ofile = "result.csv"

        if options.text is not None:
            logging.info(f"Command-line text: {options.text}")
            if ifile_list:
                return False, "Unexpected argument(s):\n\n{}".format("\n".join(ifile_list))
            self.verified_ifiles = None
        else:
            self.verified_ifiles = self.scaio.get_verified_ifile_list(ifile_list)

        self.init_kwargs = {
            "wordlist": options.wordlist,
            "tagset": options.tagset,
            "ofile": options.ofile,
            "is_stdout": options.is_stdout,
        }
        self.options = options
        return True, None

    def install_spacy(self) -> SCAProcedureResult:
        import subprocess
        from subprocess import CalledProcessError

        command = [sys.executable, "-m", "pip", "install", "-U", "spacy"]
        if get_yes_or_no(
            "Do you want to download spaCy from a Chinese mirror site? If you"
            " are inside of China, you may want to use this for a faster network"
            " connection."
        ):
            command.extend(["-i", "https://pypi.tuna.tsinghua.edu.cn/simple"])

        try:
            subprocess.run(command, check=True, capture_output=False)
        except CalledProcessError as e:
            return False, f"Failed to install spaCy: {e}"

        return True, None

    def install_model(self) -> SCAProcedureResult:
        import subprocess
        from subprocess import CalledProcessError

        if get_yes_or_no(
            "Do you want to download en_core_web_sm from sourceforge.net? If you"
            " are inside of China, you may want to use this for a faster network"
            " connection."
        ):
            command = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "https://master.dl.sourceforge.net/project/en-core-web-sm/en_core_web_sm-3.7.0-py3-none-any.whl?viasf=1",
            ]
        else:
            command = [sys.executable, "-m", "spacy", "download", "en_core_web_sm"]

        try:
            subprocess.run(command, check=True, capture_output=False)
        except CalledProcessError as e:
            return False, f"Failed to download en_core_web_sm: {e}"

        return True, None

    def check_spacy_and_model(self) -> SCAProcedureResult:
        required_version_prefix = "3.7."
        required_version_range = ">=3.7.0,<3.8.0"
        ask_install_spacy = False
        ask_install_model = False

        try:
            logging.info("Trying to load spaCy...")
            import spacy  # type: ignore # noqa: F401 'spacy' imported but unused

        except ModuleNotFoundError:
            ask_install_spacy = True
        else:
            spacy_version = spacy.__version__
            if not spacy_version.startswith(required_version_prefix):
                logging.info(
                    f"The installed version {spacy_version} of spaCy does not match the version"
                    f" required by NeoSCA: {required_version_range}"
                )
                ask_install_spacy = True
            else:
                color_print(
                    "OKGREEN",
                    "ok",
                    prefix=f"spaCy{required_version_range} has already been installed. ",
                )

        try:
            logging.info("Trying to load en_core_web_sm...")
            import en_core_web_sm  # type: ignore # noqa: F401 'en_core_web_sm' imported but unused
        except ModuleNotFoundError:
            ask_install_model = True
        else:
            model_version = en_core_web_sm.__version__
            if not model_version.startswith(required_version_prefix):
                logging.info(
                    f"The installed version {model_version} of en_core_web_sm does not match the"
                    f" version required by NeoSCA: {required_version_range}"
                )
                ask_install_model = True
            else:
                color_print(
                    "OKGREEN",
                    "ok",
                    prefix=(
                        f"en_core_web_sm{required_version_range} has already been installed. "
                    ),
                )

        if ask_install_spacy:
            is_install = get_yes_or_no(
                f"\nRunning LCA requires spaCy{required_version_range}, do you want me to"
                " install/update it for you?"
            )
            if is_install:
                self.install_spacy()
            else:
                logging.info("\nYou need to manually install it using:\npip install -U spacy")

        if ask_install_model:
            is_install = get_yes_or_no(
                f"\nRunning LCA requires spaCy's model en_core_web_sm{required_version_range},"
                " do you want me to install it for you?"
            )
            if is_install:
                self.install_model()
            else:
                logging.info(
                    "\nYou need to manually install it using:"
                    "\npython -m spacy download en_core_web_sm"
                )

        return True, None

    def exit_routine(self) -> None:
        if self.options.is_quiet or self.options.is_stdout:
            return

        color_print(
            "OKGREEN",
            f"{os_path.abspath(self.options.ofile)}",
            prefix="Output has been saved to ",
            postfix=". Done.",
        )

    def run_tmpl(func: Callable):  # type:ignore
        def wrapper(self, *args, **kwargs):
            sucess, err_msg = self.check_spacy_and_model()
            if not sucess:
                return sucess, err_msg
            if not self.options.is_stdout:
                sucess, err_msg = SCAIO.is_writable(self.options.ofile)
                if not sucess:
                    return sucess, err_msg
            func(self, *args, **kwargs)
            self.exit_routine()
            return True, None

        return wrapper

    @run_tmpl
    def run_on_text(self) -> SCAProcedureResult:
        analyzer = LCA(**self.init_kwargs)
        analyzer.analyze(text=self.options.text)
        return True, None

    @run_tmpl
    def run_on_ifiles(self) -> SCAProcedureResult:
        analyzer = LCA(**self.init_kwargs)
        analyzer.analyze(ifiles=self.verified_ifiles)
        return True, None

    def run(self) -> SCAProcedureResult:
        if self.options.text is not None:
            return self.run_on_text()
        elif self.verified_ifiles:
            return self.run_on_ifiles()
        else:
            self.args_parser.print_help()
            return True, None


def lca_main() -> None:
    ui = LCAUI()
    success, err_msg = ui.parse_args(sys.argv)
    if not success:
        logging.critical(err_msg)
        sys.exit(1)
    success, err_msg = ui.run()
    if not success:
        logging.critical(err_msg)
        sys.exit(1)



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import logging
from math import log, sqrt
import os.path as os_path
import random
import string
import sys
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

from ..scaio import SCAIO
from ..util import SCAProcedureResult


class LCA:
    def __init__(
        self,
        wordlist: str = "bnc",
        tagset: str = "ud",
        ofile: str = "result.csv",
        is_stdout: bool = False,
    ) -> None:
        self.ofile = ofile
        self.is_stdout = is_stdout

        self.scaio = SCAIO()
        self.nlp_spacy: Optional[Callable] = None

        assert wordlist in ("bnc", "anc")
        logging.debug(f"Using {wordlist.upper()} wordlist")
        self.wordlist = wordlist

        wordlist_datafile_map = {
            "bnc": "bnc_all_filtered.pickle.lzma",
            "anc": "anc_all_count.pickle.lzma",
        }

        assert tagset in ("ud", "ptb")
        logging.debug(f"Using {tagset.upper()} POS tagset")
        self.tagset = tagset

        self.get_lemma_and_pos = {
            "ud": self.get_lemma_and_udpos,
            "ptb": self.get_lemma_and_ptbpos,
        }[tagset]

        tagset_conds_map = {
            "ud": {
                "is_misc": self._is_misc_ud,
                "is_sword": self._is_sword_ud,
                "is_noun": self._is_noun_ud,
                "is_adj": self._is_adj_ud,
                "is_adv": self._is_adv_ud,
                "is_verb": self._is_verb_ud,
            },
            "ptb": {
                "is_misc": self._is_misc_ptb,
                "is_sword": self._is_sword_ptb,
                "is_noun": self._is_noun_ptb,
                "is_adj": self._is_adj_ptb,
                "is_adv": self._is_adv_ptb,
                "is_verb": self._is_verb_ptb,
            },
        }
        self.condition_map = tagset_conds_map[tagset]

        data_dir = os_path.join(os_path.dirname(os_path.dirname(__file__)), "data")
        datafile = os_path.join(data_dir, wordlist_datafile_map[wordlist])
        logging.debug(f"Loading {datafile}...")
        data = self.scaio.load_pickle_lzma_file(datafile)

        word_dict = data["word_dict"]
        adj_dict = data["adj_dict"]
        verb_dict = data["verb_dict"]
        noun_dict = data["noun_dict"]
        word_ranks = self._sort_by_value(word_dict)
        easy_words = word_ranks[-2000:]

        self.word_dict = word_dict
        self.adj_dict = adj_dict
        self.verb_dict = verb_dict
        self.noun_dict = noun_dict

        self.word_ranks = word_ranks
        self.easy_words = easy_words

        # adjust minimum sample size here
        self.standard = 50

    def _is_misc_ud(self, lemma: str, pos: str) -> bool:
        if pos in ("PUNCT", "SYM", "X", "SPACE"):
            return True
        return False

    def _is_misc_ptb(self, lemma: str, pos: str) -> bool:
        if not lemma.strip():
            return True
        if pos[0] in string.punctuation:
            return True
        if pos in ("SENT", "SYM", "HYPH"):
            return True
        return False

    def _is_sword_ud(self, lemma: str, pos: str) -> bool:
        # sophisticated word
        if lemma not in self.easy_words and pos != "NUM":
            return True
        return False

    def _is_sword_ptb(self, lemma: str, pos: str) -> bool:
        if lemma not in self.easy_words and pos != "CD":
            return True
        return False

    def _is_noun_ud(self, lemma: str, pos: str) -> bool:
        # |UD    |PTB     |
        # |------|--------|
        # |NOUN  |NN, NNS |
        # |PROPN |NNP,NNPS|
        if pos in ("NOUN", "PROPN"):
            return True
        return False

    def _is_noun_ptb(self, lemma: str, pos: str) -> bool:
        if pos.startswith("N"):
            return True
        return False

    def _is_adj_ud(self, lemma: str, pos: str) -> bool:
        if pos == "ADJ":
            return True
        return False

    def _is_adj_ptb(self, lemma: str, pos: str) -> bool:
        if pos.startswith("J"):
            return True
        return False

    def _is_adv_ud(self, lemma: str, pos: str) -> bool:
        if pos != "ADV":
            return False
        if lemma in self.adj_dict:
            return True
        if lemma.endswith("ly") and lemma[:-2] in self.adj_dict:
            return True
        return False

    def _is_adv_ptb(self, lemma: str, pos: str) -> bool:
        if not pos.startswith("R"):
            return False
        if lemma in self.adj_dict:
            return True
        if lemma.endswith("ly") and lemma[:-2] in self.adj_dict:
            return True
        return False

    def _is_verb_ud(self, lemma: str, pos: str) -> bool:
        # Don't have to filter auxiliary verbs, because the VERB tag covers
        #  main verbs (content verbs) but it does not cover auxiliary verbs
        #  and verbal copulas (in the narrow sense), for which there is the
        #  AUX tag.
        #  https://universaldependencies.org/u/pos/VERB.html
        if pos == "VERB":
            return True
        return False

    def _is_verb_ptb(self, lemma: str, pos: str) -> bool:
        if not pos.startswith("V"):
            return False
        if lemma in ("be", "have"):
            return False
        return True

    def _sort_by_value(self, d):
        """Returns the keys of dictionary d sorted by their values"""
        items = d.items()
        backitems = [[v[1], v[0]] for v in items]
        backitems.sort()
        return [backitems[i][1] for i in range(0, len(backitems))]

    def get_ndw_first_z(self, z, lemma_lst):
        """NDW for first z words in a sample"""
        ndw_first_z_types = {}
        for lemma in lemma_lst[:z]:
            ndw_first_z_types[lemma] = 1
        return len(ndw_first_z_types)

    def get_ndw_erz(self, z, lemma_lst):
        """NDW expected random z words, 10 trials"""
        ndw_erz = 0
        for _ in range(10):
            erz_lemma_lst = random.sample(lemma_lst, z)

            ndw_erz_types = set(erz_lemma_lst)
            ndw_erz += len(ndw_erz_types)
        return ndw_erz / 10

    def get_ndw_esz(self, z, lemma_lst):
        """NDW expected random sequences of z words, 10 trials"""
        ndw_esz = 0
        for _ in range(10):
            start_word = random.randint(0, len(lemma_lst) - z)
            esz_lemma_lst = lemma_lst[start_word : start_word + z]

            ndw_esz_types = set(esz_lemma_lst)
            ndw_esz += len(ndw_esz_types)
        return ndw_esz / 10

    def get_msttr(self, z, lemma_lst):
        sample_nr = 0
        msttr = 0
        while len(lemma_lst) >= z:
            sample_nr += 1
            msttr_types = {}
            for lemma in lemma_lst[:z]:
                msttr_types[lemma] = 1
            msttr += len(msttr_types) / z if z else 0
            lemma_lst = lemma_lst[z:]
        return msttr / sample_nr if sample_nr else 0

    def _safe_div(self, n1: Union[int, float], n2: Union[int, float]) -> float:
        return n1 / n2 if n2 else 0

    def compute(
        self,
        word_count_map,
        sword_count_map,
        lex_count_map,
        slex_count_map,
        verb_count_map,
        sverb_count_map,
        adj_count_map,
        adv_count_map,
        noun_count_map,
        lemma_lst,
    ):
        word_type_nr = len(word_count_map)
        word_token_nr = sum(count for count in word_count_map.values())
        lemma_nr = word_token_nr

        sword_type_nr = len(sword_count_map)
        sword_token_nr = sum(count for count in sword_count_map.values())

        lex_type_nr = len(lex_count_map)
        lex_token_nr = sum(count for count in lex_count_map.values())

        slex_type_nr = len(slex_count_map)
        slex_token_nr = sum(count for count in slex_count_map.values())

        verb_type_nr = len(verb_count_map)
        verb_token_nr = sum(count for count in verb_count_map.values())

        sverb_type_nr = len(sverb_count_map)
        # sverb_token_nr = sum(count for count in sverb_count_map.values())

        adj_type_nr = len(adj_count_map)
        # adj_token_nr = sum(count for count in adj_count_map.values())

        adv_type_nr = len(adv_count_map)
        # adv_token_nr = sum(count for count in adv_count_map.values())

        noun_type_nr = len(noun_count_map)
        noun_token_nr = sum(count for count in noun_count_map.values())

        # 1. lexical density
        ld = self._safe_div(lex_token_nr, word_token_nr)

        # 2. lexical sophistication
        # 2.1 lexical sophistication
        ls1 = self._safe_div(slex_token_nr, lex_token_nr)
        ls2 = self._safe_div(sword_type_nr, word_type_nr)

        # 2.2 verb sophistication
        vs1 = self._safe_div(sverb_type_nr, verb_token_nr)
        vs2 = self._safe_div((sverb_type_nr**2), verb_token_nr)
        cvs1 = self._safe_div(sverb_type_nr, sqrt(2 * verb_token_nr))

        # 3 lexical diversity or variation

        # 3.1 NDW, may adjust the values of "self.standard"
        ndw = ndwz = ndwerz = ndwesz = word_type_nr
        if lemma_nr >= self.standard:
            ndwz = self.get_ndw_first_z(self.standard, lemma_lst)
            ndwerz = self.get_ndw_erz(self.standard, lemma_lst)
            ndwesz = self.get_ndw_esz(self.standard, lemma_lst)

        # 3.2 TTR
        msttr = ttr = self._safe_div(word_type_nr, word_token_nr)
        if lemma_nr >= self.standard:
            msttr = self.get_msttr(self.standard, lemma_lst)
        cttr = self._safe_div(word_type_nr, sqrt(2 * word_token_nr))
        rttr = self._safe_div(word_type_nr, sqrt(word_token_nr))
        logttr = self._safe_div(log(word_type_nr), log(word_token_nr))
        uber = self._safe_div(
            log(word_token_nr, 10) * log(word_token_nr, 10),
            log(self._safe_div(word_token_nr, word_type_nr), 10),
        )

        # 3.3 verb diversity
        vv1 = self._safe_div(verb_type_nr, verb_token_nr)
        svv1 = self._safe_div(verb_type_nr * verb_type_nr, verb_token_nr)
        cvv1 = self._safe_div(verb_type_nr, sqrt(2 * verb_token_nr))

        # 3.4 lexical diversity
        lv = self._safe_div(lex_type_nr, lex_token_nr)
        vv2 = self._safe_div(verb_type_nr, lex_token_nr)
        nv = self._safe_div(noun_type_nr, noun_token_nr)
        adjv = self._safe_div(adj_type_nr, lex_token_nr)
        advv = self._safe_div(adv_type_nr, lex_token_nr)
        modv = self._safe_div((adv_type_nr + adj_type_nr), lex_token_nr)

        return (
            word_type_nr,
            sword_type_nr,
            lex_type_nr,
            slex_type_nr,
            word_token_nr,
            sword_token_nr,
            lex_token_nr,
            slex_token_nr,
            ld,
            ls1,
            ls2,
            vs1,
            vs2,
            cvs1,
            ndw,
            ndwz,
            ndwerz,
            ndwesz,
            ttr,
            msttr,
            cttr,
            rttr,
            logttr,
            uber,
            lv,
            vv1,
            svv1,
            cvv1,
            vv2,
            nv,
            adjv,
            advv,
            modv,
        )

    def _analyze(
        self,
        *,
        filepath: Optional[str] = None,
        text: Optional[str] = None,
    ):
        assert (not filepath) ^ (not text)

        if filepath is not None:
            logging.info(f"Processing {filepath}...")
            text = self.scaio.read_file(filepath)

        if text is None:
            return None

        condition_map = self.condition_map

        word_count_map: Dict[str, int] = {}
        sword_count_map: Dict[str, int] = {}
        lex_count_map: Dict[str, int] = {}
        slex_count_map: Dict[str, int] = {}
        verb_count_map: Dict[str, int] = {}
        sverb_count_map: Dict[str, int] = {}
        adj_count_map: Dict[str, int] = {}
        adv_count_map: Dict[str, int] = {}
        noun_count_map: Dict[str, int] = {}
        lemma_lst = []

        g = self.get_lemma_and_pos(text)

        # Universal POS tags: https://universaldependencies.org/u/pos/
        for lemma, pos in g:
            if condition_map["is_misc"](lemma, pos):
                continue

            is_sophisticated = False
            is_lexical = False
            is_verb = False

            word_count_map[lemma] = word_count_map.get(lemma, 0) + 1
            lemma_lst.append(lemma)

            if condition_map["is_noun"](lemma, pos):
                noun_count_map[lemma] = noun_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a noun')

                lex_count_map[lemma] = lex_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a lexical word')

                is_lexical = True

            elif condition_map["is_adj"](lemma, pos):
                adj_count_map[lemma] = adj_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as an adjective')

                lex_count_map[lemma] = lex_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a lexical word')

                is_lexical = True

            elif condition_map["is_adv"](lemma, pos):
                adv_count_map[lemma] = adv_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as an adverb')

                lex_count_map[lemma] = lex_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a lexical word')

                is_lexical = True

            elif condition_map["is_verb"](lemma, pos):
                verb_count_map[lemma] = verb_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a verb')

                lex_count_map[lemma] = lex_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a lexical word')

                is_lexical = True
                is_verb = True

            if condition_map["is_sword"](lemma, pos):
                sword_count_map[lemma] = sword_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a sophisticated word')

                is_sophisticated = True

            if is_lexical and is_sophisticated:
                slex_count_map[lemma] = slex_count_map.get(lemma, 0) + 1
                logging.debug(f'Counted "{lemma}" as a sophisticated lexical word')
                if is_verb:
                    sverb_count_map[lemma] = sverb_count_map.get(lemma, 0) + 1
                    logging.debug(f'Counted "{lemma}" as a sophisticated verb')

        return self.compute(
            word_count_map,
            sword_count_map,
            lex_count_map,
            slex_count_map,
            verb_count_map,
            sverb_count_map,
            adj_count_map,
            adv_count_map,
            noun_count_map,
            lemma_lst,
        )

    def analyze(
        self, *, ifiles: Optional[List[str]] = None, text: Optional[str] = None
    ) -> SCAProcedureResult:
        if not (ifiles is None) ^ (text is None):
            return False, "One and only one of (input files, text) should be given."

        import csv

        handle = (
            open(self.ofile, "w", encoding="utf-8", newline="")
            if not self.is_stdout
            else sys.stdout
        )
        fieldnames = (
            "filename",
            "wordtypes (word types)",
            "swordtypes (sophisticated word types)",
            "lextypes (lexical types)",
            "slextypes (sophisticated lexical types)",
            "wordtokens (word tokens)",
            "swordtokens (sophisticated word tokens)",
            "lextokens (lexical tokens)",
            "slextokens (sophisticated lexical tokens)",
            "LD (lexical density)",
            "LS1 (lexical sophistication-I)",
            "LS2 (lexical sophistication-II)",
            "VS1 (verb sophistication-I)",
            "VS2 (verb sophistication-II)",
            "CVS1 (corrected VS1)",
            "NDW (number of different words)",
            "NDW-50 (NDW, first 50 words)",
            "NDW-ER50 (NDW, expected random 50)",
            "NDW-ES50 (NDW, expected sequence 50)",
            "TTR (type-token ratio)",
            "MSTTR (mean segmental TTR, 50)",
            "CTTR (corrected TTR)",
            "RTTR (root TTR)",
            "LogTTR (bilogarithmic TTR)",
            "Uber (Uber Index)",
            "LV (lexical word variation)",
            "VV1 (verb variation-I)",
            "SVV1 (squared VV1)",
            "CVV1 (corrected VV1)",
            "VV2 (verb variation-II)",
            "NV (noun variation)",
            "AdjV (adjective variation)",
            "AdvV (adverb variation)",
            "ModV (modifier variation)",
        )
        csv_writer = csv.writer(handle)
        csv_writer.writerow(fieldnames)

        if text is not None:
            values = self._analyze(text=text)
            if values is not None:
                values = [str(round(v, 4)) for v in values]
                values.insert(0, "cmdline_text")
                csv_writer.writerow(values)

        else:
            for ifile in ifiles:  # type: ignore
                values = self._analyze(filepath=ifile)
                if values is not None:
                    values = [str(round(v, 4)) for v in values]
                    values.insert(0, ifile)
                    csv_writer.writerow(values)

        handle.close()

        return True, None

    def ensure_spacy_initialized(func: Callable):  # type:ignore
        def wrapper(self, *args, **kwargs):
            if self.nlp_spacy is None:
                logging.info("Initializing spaCy...")
                import spacy  # type: ignore

                # default spacy pipeline: "tok2vec", "tagger", "parser",
                #  "attribute_ruler", "lemmatizer", "ner"
                self.nlp_spacy = spacy.load("en_core_web_sm", exclude=["ner", "parser"])

            return func(self, *args, **kwargs)

        return wrapper

    @ensure_spacy_initialized
    def get_lemma_and_udpos(self, text: str) -> Generator[Tuple[str, str], Any, None]:
        doc = self.nlp_spacy(text)  # type:ignore
        for token in doc:
            yield (token.lemma_.lower(), token.pos_)

    @ensure_spacy_initialized
    def get_lemma_and_ptbpos(self, text: str) -> Generator[Tuple[str, str], Any, None]:
        doc = self.nlp_spacy(text)  # type:ignore
        for token in doc:
            yield (token.lemma_.lower(), token.tag_)



#!/usr/bin/env python3
# -*- coding=utf-8 -*-
import logging
import sys
from typing import Optional

import jpype
from jpype import JClass


class StanfordParser:
    def __init__(
        self,
        classpaths: Optional[list] = None,
        nthreads: int = 1,
        # tested on a 16kb file: 3m23s with 2 threads vs. 3m21s with 1 threads
        max_memory: str = "3072m",  # 3g
    ) -> None:
        self.classpaths = classpaths if classpaths is not None else []
        self.nthreads = nthreads
        self.max_memory = max_memory

        self.PARSER_GRAMMAR = "edu.stanford.nlp.parser.lexparser.LexicalizedParser"
        self.PARSER_MODEL = "edu/stanford/nlp/models/lexparser/englishPCFG.ser.gz"
        self.DOCUMENT_PREPROCESSOR = "edu.stanford.nlp.process.DocumentPreprocessor"
        self.parsed_sent_num = 0
        self.long_sent_num = 0
        self.no_parse_num = 0
        self.PROMPT_PARSING = " Parsing [sent. {} len. {}]: {}"
        self.PROMPT_LONG_SENTENCE = "Sentence longer than {}. Skipping: {}\n"
        self.PROMPT_NO_PARSE = (
            'Sentence has no parse using PCFG grammar (or no PCFG fallback). Skipping: "{}"'
        )
        self.PROMPT_LONG_SENTENCE_SUMMARY = (
            "{} sentences are skipped because they are longer than {}.\n"
        )
        self.PROMPT_NO_PARSE_SUMMARY = (
            "{} sentences are skipped because they have no parse using PCFG grammar (or no PCFG"
            " fallback).\n"
        )
        self.init_parser()

    def init_parser(self):
        if not jpype.isJVMStarted():
            logging.debug("starting JVM...")
            # Note that isJVMStarted may be renamed to isJVMRunning in the future.
            # In jpype's _core.py:
            # > TODO This method is horribly named.  It should be named isJVMRunning as
            # > isJVMStarted would seem to imply that the JVM was started at some
            # > point without regard to whether it has been shutdown.
            jpype.startJVM(f"-Xmx{self.max_memory}", classpath=self.classpaths)

        LexicalizedParser = JClass(self.PARSER_GRAMMAR)
        RedwoodConfiguration = JClass("edu.stanford.nlp.util.logging.RedwoodConfiguration")
        RedwoodConfiguration.empty().capture(jpype.java.lang.System.err).apply()
        self.lp = LexicalizedParser.loadModel(self.PARSER_MODEL)
        options = ["-outputFormat", "penn", "-nthreads", str(self.nthreads)]
        self.lp.setOptionFlags(options)

    def _is_long(self, sentence_length: int, max_length: Optional[int] = None) -> bool:
        if max_length is not None and max_length < sentence_length:
            return True
        return False

    def parse_sentence(self, sentence, max_length: Optional[int] = None) -> str:
        """Parse a single sentence"""
        sentence_length = len(sentence)
        plain_sentence = " ".join(str(w.toString()) for w in sentence)
        if self._is_long(sentence_length, max_length):  # pragma: no cover
            self.long_sent_num += 1
            sys.stderr.write(
                self.PROMPT_LONG_SENTENCE.format(max_length, plain_sentence)
            )  # pragma: no cover
            return ""
        self.parsed_sent_num += 1
        logging.info(
            self.PROMPT_PARSING.format(self.parsed_sent_num, sentence_length, plain_sentence)
        )
        parse = self.lp.apply(sentence)
        if parse is not None:
            return str(parse.pennString().replaceAll("\r", ""))
        else:  # pragma: no cover
            self.no_parse_num += 1
            logging.warning(self.PROMPT_NO_PARSE.format(plain_sentence))
            return ""

    def parse_text(
        self, text: str, max_length: Optional[int] = None, is_pretokenized: bool = False
    ) -> str:
        doc = JClass(self.DOCUMENT_PREPROCESSOR)(jpype.java.io.StringReader(text))
        if is_pretokenized:
            WhitespaceTokenizerFactory = JClass(
                "edu.stanford.nlp.process.WhitespaceTokenizer$WhitespaceTokenizerFactory"
            )
            doc.setTokenizerFactory(WhitespaceTokenizerFactory.newTokenizerFactory())
        trees = "\n".join(self.parse_sentence(sentence, max_length) for sentence in doc)
        return trees

    def refresh_counters(self, max_length: Optional[int] = None) -> None:
        if max_length is not None and self.long_sent_num > 0:
            sys.stderr.write(  # pragma: no cover
                self.PROMPT_LONG_SENTENCE_SUMMARY.format(self.long_sent_num, max_length)
            )
        if self.no_parse_num > 0:
            sys.stderr.write(
                self.PROMPT_NO_PARSE_SUMMARY.format(self.no_parse_num)
            )  # pragma: no cover

        self.long_sent_num = 0
        self.parsed_sent_num = 0
        self.no_parse_num = 0

    def parse(
        self,
        text: str,
        max_length: Optional[int] = None,
        newline_break: str = "never",
        is_pretokenized: bool = False,
        is_reserve_parsed: bool = False,
        ofile_parsed: str = "cmdline_text.parsed",
        is_stdout: bool = False,
    ) -> str:
        assert newline_break in ("never", "always", "two")
        if newline_break == "never":
            trees = self.parse_text(text, max_length, is_pretokenized)
        else:
            if newline_break == "always":
                paragraphs = list(filter(None, text.split("\n")))
            else:
                import re

                paragraphs = re.split(r"(?:\r\n|\n|\r){2,}", text)
            trees = "\n".join(
                self.parse_text(paragraph, max_length, is_pretokenized)
                for paragraph in paragraphs
            )
        if is_reserve_parsed:
            if not is_stdout:
                with open(ofile_parsed, "w", encoding="utf-8") as f:
                    f.write(trees)
            else:
                sys.stdout.write(trees + "\n")
        self.refresh_counters(max_length)
        return trees



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import os
import os.path as os_path
import sys

IS_WINDOWS = sys.platform == "win32"
IS_MAC = sys.platform == "darwin"
IS_LINUX = sys.platform == "linux"
user_home = os_path.expanduser("~")
if IS_WINDOWS:  # pragma: no cover
    user_name = os_path.basename(user_home)
    if user_name.isascii():
        USER_SOFTWARE_DIR = os.getenv("AppData", os_path.join(user_home, "AppData", "Roaming"))
    else:  # if user_name contains non-latin chars
        USER_SOFTWARE_DIR = os_path.splitdrive(user_home)[0] + os_path.sep  # C:\
else:
    USER_SOFTWARE_DIR = os_path.join(user_home, ".local", "share")



import setuptools

with open("./README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("./neosca/about.py", "r", encoding="utf-8") as f:
    about = {}
    exec(f.read(), about)

setuptools.setup(
    name="neosca",
    version=about["__version__"],
    author="Long Tan",
    author_email="tanloong@foxmail.com",
    url="https://github.com/tanloong/neosca",
    packages=["neosca", "neosca.lca"],
    package_data={
        "neosca": [
            "data/structure_data.json",
            "data/anc_all_count.pickle.lzma",
            "data/bnc_all_filtered.pickle.lzma",
        ]
    },
    description="Another syntactic complexity analyzer of written English language samples",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["JPype1", "charset-normalizer"],
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["nsca = neosca:main", "nsca-lca = neosca:lca_main", "nsca-tregex = neosca:tregex_cli"]},
)



import argparse
import glob
import logging
import os
import os.path as os_path
import sys
from typing import Callable, List, Optional

from .about import __version__
from .scaenv import (
    JAVA_HOME,
    STANFORD_PARSER_HOME,
    STANFORD_TREGEX_HOME,
    getenv,
    search_java_home,
    setenv,
)
from .scaio import SCAIO
from .scaprint import color_print
from .util import SCAProcedureResult


class SCAUI:
    def __init__(self) -> None:
        self.scaio = SCAIO()
        self.supported_ifile_types = list(self.scaio.ext_read_map.keys())
        self.cwd = os.getcwd()
        self.args_parser: argparse.ArgumentParser = self.create_args_parser()
        self.options: argparse.Namespace = argparse.Namespace()

    def create_args_parser(self) -> argparse.ArgumentParser:
        args_parser = argparse.ArgumentParser(
            prog="nsca", formatter_class=argparse.RawDescriptionHelpFormatter
        )
        args_parser.add_argument(
            "--version",
            action="store_true",
            default=False,
            help="Show version and exit.",
        )
        args_parser.add_argument(
            "--list",
            dest="list_fields",
            action="store_true",
            default=False,
            help="List built-in measures.",
        )
        args_parser.add_argument(
            "--expand-wildcards",
            dest="expand_wildcards",
            action="store_true",
            default=False,
            help=(
                "Print all files that match your wildcard pattern. This can help you ensure that"
                " your pattern matches all desired files and excludes any unwanted ones. Note"
                " that files that do not exist on the computer will not be included in the"
                " output, even if they match the specified pattern."
            ),
        )
        args_parser.add_argument(
            "--max-length",
            metavar="<max_length>",
            dest="max_length",
            type=int,
            default=None,
            help=(
                "Set the longest sentence to parse (inclusively). Sentences longer than"
                " <max_length> will be skipped, with a message printed to stderr. When this is"
                " not specified, the program will try to analyze sentences of any lengths, but"
                " may run out of memory trying to do so"
                " (https://nlp.stanford.edu/software/parser-faq.html#k)."
            ),
        )
        args_parser.add_argument(
            "--newline-break",
            dest="newline_break",
            choices=["never", "always", "two"],
            default="never",
            help=(
                "Whether to treat newlines as sentence breaks. This option has 3 legal values."
                ' "never" means to ignore newlines for the purpose of sentence splitting, and is'
                " appropriate for continuous text with hard line breaks when just the"
                " non-whitespace characters should be used to determine sentence breaks."
                ' "always" means to treat a newline as a sentence break, but there still may be'
                ' more than one sentences per line. "two" means to take two or more consecutive'
                " newlines as a sentence break, and is for text with hard line breaks and a"
                ' blank line between paragraphs. The default is "never".'
            ),
        )
        args_parser.add_argument(
            "--combine-files",
            "-c",
            metavar="<subfile>",
            dest="subfiles_list",
            action="append",
            default=None,
            nargs="+",
            help="Combine frequency output of multiple files.",
        )
        args_parser.add_argument(
            "--text",
            "-t",
            metavar="<text>",
            default=None,
            help="Pass text through the command line.",
        )
        args_parser.add_argument(
            "--ftype",
            dest="ifile_types",
            choices=self.supported_ifile_types,
            default=self.supported_ifile_types,
            nargs="+",
            help=(
                "Analyze files of the specified type(s). If not set, the program will process"
                " files of all supported types."
            ),
        )
        args_parser.add_argument(
            "--pretokenized",
            dest="is_pretokenized",
            action="store_true",
            default=False,
            help="Assume that the text has already been tokenized.",
        )
        args_parser.add_argument(
            "--output-file",
            "-o",
            metavar="<filename>",
            dest="ofile_freq",
            default=None,
            help='Specify an output file. The default is "result.csv".',
        )
        args_parser.add_argument(
            "--output-format",
            dest="oformat_freq",
            choices=["csv", "json"],
            default="csv",
            help='Output format, the default is "csv".',
        )
        args_parser.add_argument(
            "--stdout",
            dest="is_stdout",
            action="store_true",
            default=False,
            help="Write the frequency output to the stdout instead of saving it to a file.",
        )
        args_parser.add_argument(
            "--select",
            metavar="<measure>",
            dest="selected_measures",
            default=None,
            nargs="+",
            help=(
                "Select only some of the measures to analyze. The builtin measures include"
                ' "W", "S", "VP", "C", "T", "DC", "CT", "CP", "CN", "MLS", "MLT", "MLC", "C/S",'
                ' "VP/T", "C/T", "DC/C", "DC/T", "T/S", "CT/T", "CP/T", "CP/C", "CN/T", and'
                ' "CN/C".'
            ),
        )
        args_parser.add_argument(
            "--reserve-parsed",
            "-p",
            dest="is_reserve_parsed",
            action="store_true",
            default=False,
            help="Reserve the parse trees produced by the Stanford Parser.",
        )
        args_parser.add_argument(
            "--reserve-matched",
            "-m",
            dest="is_reserve_matched",
            default=False,
            action="store_true",
            help="Reserve the matched subtrees produced by the Stanford Tregex.",
        )
        args_parser.add_argument(
            "--no-parse",
            dest="is_skip_parsing",
            action="store_true",
            default=False,
            help=(
                "Assume input as parse trees. By default, the program expects"
                " raw text as input that will be parsed before querying. If you"
                " already have parsed input files, use this flag to indicate that"
                " the program should skip the parsing step and proceed directly"
                " to querying. When this flag is set, the is_skip_querying and"
                " reserve_parsed are automatically set as False."
            ),
        )
        args_parser.add_argument(
            "--no-query",
            dest="is_skip_querying",
            action="store_true",
            default=False,
            help="Parse the input files, save the parse trees and exit.",
        )
        args_parser.add_argument(
            "--quiet",
            dest="is_quiet",
            action="store_true",
            default=False,
            help="Stop NeoSCA from printing anything except for final results.",
        )
        args_parser.add_argument(
            "--verbose",
            dest="is_verbose",
            action="store_true",
            default=False,
            help="Print detailed logging messages.",
        )
        args_parser.add_argument(
            "--check-depends",
            dest="check_depends",
            action="store_true",
            default=False,
            help=(
                "Check and install NeoSCA's dependencies: Java, Stanford Parser, and"
                " Stanford Tregex."
            ),
        )
        args_parser.add_argument(
            "--yes",
            dest="is_assume_yes",
            action="store_true",
            default=False,
            help="Assume the answer to all prompts is yes, used when installing dependencies.",
        )
        args_parser.add_argument(
            "--config",
            dest="config",
            default=None,
            help=(
                "Use custom json file where you can define your own syntactic structures to"
                " search or calculate."
            ),
        )
        args_parser.epilog = """Examples:
1. nsca sample1.txt
2. nsca "sample 1.txt"
    Filenames containing whitespace should be quoted.
3. nsca sample1.txt -o sample1.csv
    Customize the output filename instead of the default result.csv.
4. nsca sample1.txt -p
    Reserve parsing results generated by Stanford Parser.
5. nsca sample1.txt -m
    Reserve matching results generated by Stanford Tregex.
6. nsca samples/
    Analyze every file with .txt extension located in the "samples/" directory
7. nsca sample1.txt sample2.txt
    Analyze a list of input files.
8. nsca sample*.txt
    Analyze files with name starting with "sample" and ending with ".txt"
9. nsca sample10[1-9].txt sample1[1-9][0-9].txt sample200.txt
    Analyze files ranging from sample101.txt to sample200.txt.
10. nsca sample10[1-9].txt sample1[1-9][0-9].txt sample200.txt --expand-wildcards
    Expand the specified wildcards and exit.
11. nsca sample1.txt --max-length 100
    Only analyze sentences with lengths shorter than or equal to 100.
12. nsca sample1.txt --newline-break always
    Consider newlines as sentence breaks.
13. nsca --select VP T DC/C -- sample1.txt
    Select a subset of measures to analyze. Use -- to separate input
    filenames from the selected measures, or otherwise the program will take
    "sample1.txt" as a measure and then raise an error. Arguments other than
    input filenames should be specified at the left side of --.
14. nsca -c sample1-sub1.txt sample1-sub2.txt
    Add up frequencies of the 9 syntactic structures of the subfiles and compute
    values of the 14 syntactic complexity indices for the imaginary parent file.
15. nsca -c samples/
    Combine all the input files within the "samples/" directory
16. nsca -c sample1-sub*.txt
    Wildcards are supported for -c.
17. nsca -c sample1-sub*.txt -c sample2-sub*.txt
    Use multiple -c to combine different lists of subfiles respectively.
18. nsca -c sample1-sub*.txt -c sample2-sub*.txt -- sample[3-9].txt
    Use -- to separate input filenames from names of the subfiles.
19. nsca sample1.txt --no-query
    Parse the input files, save the parsed trees and exit.
20. nsca sample1.parsed --no-parse
    Assume input as parse trees. Skip the parsing step and proceed directly to querying.
21. nsca --config nsca.json sample1.txt
    Use nsca.json where you can defined your own syntactic structures to search or calculate.

Contact:
1. https://github.com/tanloong/neosca/issues
2. tanloong@foxmail.com
"""
        return args_parser

    def parse_args(self, argv: List[str]) -> SCAProcedureResult:
        idx: Optional[int] = None
        if "--" in argv[1:]:
            idx = argv.index("--")
        if idx is not None:
            options, ifile_list = self.args_parser.parse_args(argv[1:idx]), argv[idx + 1 :]
        else:
            options, ifile_list = self.args_parser.parse_known_args(argv[1:])

        assert not (
            options.is_quiet and options.is_verbose
        ), "logging cannot be quiet and verbose at the same time"
        if options.is_quiet:
            logging.basicConfig(format="%(message)s", level=logging.CRITICAL)
        elif options.is_verbose:
            logging.basicConfig(format="%(message)s", level=logging.DEBUG)
        else:
            logging.basicConfig(format="%(message)s", level=logging.INFO)

        if options.is_skip_querying:
            options.is_reserve_parsed = True
        if options.is_skip_parsing:
            options.is_skip_querying = False
            options.is_reserve_parsed = False

        if options.text is not None:
            if ifile_list:
                return False, "Unexpected argument(s):\n\n{}".format("\n".join(ifile_list))
            logging.info(f"Command-line text: {options.text}")
            self.verified_ifiles = None
        else:
            self.verified_ifiles = self.scaio.get_verified_ifile_list(ifile_list)

        if options.subfiles_list is None:
            self.verified_subfiles_list: List[list] = []
        else:
            verified_subfiles_list = []
            for subfiles in options.subfiles_list:
                verified_subfiles = []
                for path in subfiles:
                    if os_path.isfile(path):
                        verified_subfiles.append(path)
                    elif os_path.isdir(path):
                        for ftype in options.ifile_types:
                            verified_subfiles.extend(glob.glob(f"{path}{os_path.sep}*.{ftype}"))
                    elif glob.glob(path):
                        verified_subfiles.extend(glob.glob(path))
                    else:
                        return False, f"No such file as\n\n{path}"
                if len(verified_subfiles) == 1:
                    logging.critical(
                        f"Only 1 subfile provided: ({verified_subfiles[0]}). There should be 2"
                        " or more subfiles to combine."
                    )
                    sys.exit(1)
                verified_subfiles_list.append(verified_subfiles)
            self.verified_subfiles_list = verified_subfiles_list

        self.odir_matched = "result_matches"
        if options.ofile_freq is not None:
            self.odir_matched = os_path.splitext(options.ofile_freq)[0] + "_matches"
            ofile_freq_ext = os_path.splitext(options.ofile_freq)[-1].lstrip(".")
            if ofile_freq_ext not in ("csv", "json"):
                return (
                    False,
                    (
                        f"The file extension {ofile_freq_ext} is not supported. Use one of"
                        " the following:\n1. csv\n2. json"
                    ),
                )
            if ofile_freq_ext != options.oformat_freq:
                options.oformat_freq = ofile_freq_ext
        else:
            options.ofile_freq = "result." + options.oformat_freq

        is_max_length_given_and_lt_zero = (
            options.max_length is not None and options.max_length < 0
        )
        if is_max_length_given_and_lt_zero or options.max_length == 0:
            return False, 'The value of "--max-length" should be greater than 0.'

        if options.selected_measures is not None:
            # Drop duplicates while retain order. Starting from Python 3.7, the
            # built-in dictionary is guaranteed to maintain the insertion order
            options.selected_measures = list(dict.fromkeys(options.selected_measures))

        user_config = options.config
        if user_config is not None:
            if not os_path.isfile(user_config):
                return False, f"no such file as\n\n{user_config}"
            if not user_config.endswith(".json"):
                return False, f'"{user_config}" does not seem like a json file.'
            logging.debug(f"[Main] Using configuration file {user_config}")
        else:
            default_config_file = "nsca.json"
            if os_path.isfile(default_config_file):
                user_config = default_config_file
                logging.debug(f"[Main] Using configuration file {user_config}")
            else:
                logging.debug("[Main] No configuration file found")

        self.init_kwargs = {
            "ofile_freq": options.ofile_freq,
            "oformat_freq": options.oformat_freq,
            "stanford_parser_home": "",
            "stanford_tregex_home": "",
            "odir_matched": self.odir_matched,
            "newline_break": options.newline_break,
            "max_length": options.max_length,
            "selected_measures": options.selected_measures,
            "is_reserve_parsed": options.is_reserve_parsed,
            "is_reserve_matched": options.is_reserve_matched,
            "is_stdout": options.is_stdout,
            "is_skip_querying": options.is_skip_querying,
            "is_skip_parsing": options.is_skip_parsing,
            "is_pretokenized": options.is_pretokenized,
            "config": user_config,
        }
        self.options = options
        return True, None

    def check_java(self) -> SCAProcedureResult:
        java_home = getenv(JAVA_HOME)
        if java_home is None:
            logging.debug("JAVA_HOME either does not exist or is not a valid dirname.")
            java_home = search_java_home()
            if java_home is None:
                from .depends_installer import DependsInstaller
                from .depends_installer import JAVA

                installer = DependsInstaller()
                sucess, err_msg = installer.install(
                    JAVA, is_assume_yes=self.options.is_assume_yes
                )
                if not sucess:
                    return sucess, err_msg
                else:
                    java_home = err_msg
            java_bin = os_path.join(java_home, "bin")  # type:ignore
            path_orig = os.getenv("PATH", "")
            setenv(
                "JAVA_HOME",
                [java_home],  # type:ignore
                is_override=True,
                is_quiet=self.options.is_quiet,
            )
            setenv(
                "PATH", [java_bin], is_override=False, is_quiet=self.options.is_quiet
            )  # type:ignore
            os.environ["JAVA_HOME"] = java_home  # type:ignore
            os.environ["PATH"] = java_bin + os.pathsep + path_orig  # type:ignore
        elif not self.options.is_quiet:
            color_print("OKGREEN", "ok", prefix="Java has already been installed. ")
        return True, None

    def check_stanford_parser(self) -> SCAProcedureResult:
        stanford_parser_home = getenv(STANFORD_PARSER_HOME)
        if stanford_parser_home is None:
            from .depends_installer import DependsInstaller
            from .depends_installer import STANFORD_PARSER

            installer = DependsInstaller()
            sucess, err_msg = installer.install(
                STANFORD_PARSER, is_assume_yes=self.options.is_assume_yes
            )
            if not sucess:
                return sucess, err_msg
            else:
                stanford_parser_home = err_msg
                assert stanford_parser_home is not None
                setenv(
                    STANFORD_PARSER_HOME,
                    [stanford_parser_home],
                    is_override=True,
                    is_quiet=self.options.is_quiet,
                )
        elif not self.options.is_quiet:
            color_print("OKGREEN", "ok", prefix="Stanford Parser has already been installed. ")

        self.init_kwargs.update({"stanford_parser_home": stanford_parser_home})
        return True, None

    def check_stanford_tregex(self) -> SCAProcedureResult:
        stanford_tregex_home = getenv(STANFORD_TREGEX_HOME)
        if stanford_tregex_home is None:
            from .depends_installer import DependsInstaller
            from .depends_installer import STANFORD_TREGEX

            installer = DependsInstaller()
            sucess, err_msg = installer.install(
                STANFORD_TREGEX, is_assume_yes=self.options.is_assume_yes
            )
            if not sucess:
                return sucess, err_msg
            else:
                stanford_tregex_home = err_msg
                assert stanford_tregex_home is not None
                setenv(
                    STANFORD_TREGEX_HOME,
                    [stanford_tregex_home],
                    is_override=True,
                    is_quiet=self.options.is_quiet,
                )
        elif not self.options.is_quiet:
            color_print("OKGREEN", "ok", prefix="Stanford Tregex has already been installed. ")
        self.init_kwargs.update({"stanford_tregex_home": stanford_tregex_home})
        return True, None

    def check_depends(self) -> SCAProcedureResult:
        success_java, err_msg_java = self.check_java()
        success_parser, err_msg_parser = self.check_stanford_parser()
        success_tregex, err_msg_tregex = self.check_stanford_tregex()

        sucesses = (success_java, success_parser, success_tregex)
        err_msges = (err_msg_java, err_msg_parser, err_msg_tregex)
        if all(sucesses):
            return True, None
        else:
            err_msg = "\n\n".join(
                map(lambda p: p[1] if not p[0] else "", zip(sucesses, err_msges))  # type:ignore
            )
            return False, err_msg.strip()

    def check_python(self) -> SCAProcedureResult:
        v_info = sys.version_info
        if v_info.minor >= 7 and v_info.major == 3:
            return True, None
        else:
            return (
                False,
                (
                    f"Error: Python {v_info.major}.{v_info.minor} is too old."
                    " NeoSCA only supports Python 3.7 or higher."
                ),
            )

    def exit_routine(self) -> None:
        if self.options.is_quiet or self.options.is_stdout:
            return

        msg_num = 0
        if not self.options.is_skip_querying:
            msg_num += 1
            color_print(
                "OKGREEN",
                f"{os_path.abspath(self.options.ofile_freq)}",
                prefix=f"{msg_num}. Frequency output was saved to ",
                postfix=".",
            )
        if self.options.is_reserve_parsed:
            if self.verified_ifiles or self.verified_subfiles_list:
                msg_num += 1
                logging.info(
                    f"{msg_num}. parse trees were saved corresponding to input files,"
                    ' with the same name but a ".parsed" extension.'
                )
            elif self.options.text is not None:
                msg_num += 1
                color_print(
                    "OKGREEN",
                    f"{self.cwd}{os.sep}cmdline_text.parsed",
                    prefix=f"{msg_num}. parse trees were saved to ",
                    postfix=".",
                )
        if self.options.is_reserve_matched:
            msg_num += 1
            color_print(
                "OKGREEN",
                f"{os_path.abspath(self.odir_matched)}",
                prefix=f"{msg_num}. Matched subtrees were saved to ",
                postfix=".",
            )
        if msg_num > 0:
            logging.info("Done.")

    def run_tmpl(func: Callable):  # type:ignore
        def wrapper(self, *args, **kwargs):
            sucess, err_msg = self.check_python()
            if not sucess:
                return sucess, err_msg
            sucess, err_msg = self.check_depends()
            if not sucess:
                return sucess, err_msg
            if not self.options.is_stdout:
                sucess, err_msg = SCAIO.is_writable(self.options.ofile_freq)
                if not sucess:
                    return sucess, err_msg
            func(self, *args, **kwargs)
            self.exit_routine()
            return True, None

        return wrapper

    @run_tmpl
    def run_on_text(self) -> None:
        from .neosca import NeoSCA

        analyzer = NeoSCA(**self.init_kwargs)
        analyzer.run_on_text(self.options.text)

    @run_tmpl
    def run_on_ifiles(self) -> None:
        from .neosca import NeoSCA

        analyzer = NeoSCA(**self.init_kwargs)

        analyzer.run_on_ifiles(
            files=self.verified_ifiles or [], subfiles_list=self.verified_subfiles_list or []
        )

    def run(self) -> SCAProcedureResult:
        if self.options.version:
            return self.show_version()
        elif self.options.list_fields:
            return self.list_fields()
        elif self.options.expand_wildcards:
            return self.expand_wildcards()
        elif self.options.check_depends:
            return self.check_depends()
        elif self.options.text is not None:
            return self.run_on_text()  # type: ignore
        elif self.verified_ifiles or self.verified_subfiles_list:
            return self.run_on_ifiles()  # type: ignore
        else:
            self.args_parser.print_help()
            return True, None

    def list_fields(self) -> SCAProcedureResult:
        from .structure_counter import StructureCounter

        counter = StructureCounter()
        for s_name in counter.selected_measures:
            print(f"{s_name}: {counter.get_structure(s_name).description}")
        return True, None

    def expand_wildcards(self) -> SCAProcedureResult:
        is_not_found = True
        if self.verified_ifiles:
            is_not_found = False
            print("Input files:")
            for i, ifile in enumerate(self.verified_ifiles, 1):
                print(f" {i}. {ifile}")
        if self.verified_subfiles_list:
            is_not_found = False
            for i, subfiles in enumerate(self.verified_subfiles_list, 1):
                print(f"Input subfile list {i}:")
                for j, subfile in enumerate(subfiles, 1):
                    print(f" {j}. {subfile}")
        if is_not_found:
            print("0 files and subfiles are found.")
        return True, None

    def show_version(self) -> SCAProcedureResult:
        print(__version__)
        return True, None


def main() -> None:
    ui = SCAUI()
    success, err_msg = ui.parse_args(sys.argv)
    if not success:
        logging.critical(err_msg)
        sys.exit(1)
    success, err_msg = ui.run()
    if not success:
        logging.critical(err_msg)
        sys.exit(1)
